# -*- coding: utf-8 -*-
#
# Copyright 2018 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls_S (CoMPUter-based case fiLeS -
# Deutschlandstipendium).
#
# mpuls_S is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls_S is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls_S. If not, see <http://www.gnu.org/licenses/>.

import logging

from mpulsweb.controllers.formularpage import FormularpageController as \
    MpulsFormularpageController, PageRenderer
from mpulsweb.lib.formedrenderer import ViewRenderer as MpulsViewRenderer, \
    ViewPrintRenderer as MpulsViewPrintRenderer
from mpulsweb.model.agencysettings import Agency
from formed.renderer.html import Item, ErrorRenderer, TreeFormData

from mpulsweb.lib.base import g


log = logging.getLogger(__name__)


class CoachesFormData(TreeFormData):

    """TreeFormData that handles bool-groups."""

    def evaluate_rules(self, meta, warnings_only=False):
        """Extend the inherited method to handle bool-groups correctly.

        If the meta object is neither a bool-group (as determined form
        its name) nor one the bools inside such a group (as determined
        by its name being in g.bool_group_bools) this method just
        delegates to the inherited method.

        For bool-groups it evaluates any warning rules associated with
        it (which should be all such groups in jmdcoachesweb because we
        generate such rules). We cannot just use the inherited method
        for this, because there the warnings would be ignored because
        the group itself is not a variable referenced by the rules. The
        rules only reference the bools inside the group.

        For the bools inside such groups it simply returns empty lists
        of warnings, errors and notices. The bools do have warning rules
        but we need to ignore them when rendering the form. The warning
        for the form is covered by the rule associated with the
        bool-group. The warning for the bools themselves are only needed
        for automatic phase changes.
        """
        meta_name = meta.getName()
        if meta_name in g.bool_group_bools:
            # ignore the rules associated with the checkboxes in the
            # bool groups so that here we only look at the global rule
            # for the group
            return ([], [], [])

        if not meta_name.startswith("bool-group-"):
            return super(CoachesFormData, self).evaluate_rules(meta,
                                                               warnings_only)

        context = self.get_variable_provider()
        warnings = []
        for rule in self.formed.name2markingrules.get(meta_name, ()):
            if not rule.isWarning():
                continue
            if not self.rule_available(self.formed, rule, context):
                continue
            for result, variables in rule.getExpr().evaluate_instances(context):
                if not result:
                    warnings.append(rule)

        return (warnings, [], [])


class ViewRenderer(MpulsViewRenderer):

    def has_warnings(self, nc):
        """Extend the inherited method with special support for bool-groups.

        This method treats bools in bool-groups special: It returns true
        iff the containing bool-group has a warning.

        Other cases are delegated to the inherited method.
        """
        bool_group = g.bool_group_bools.get(nc.getName())
        if bool_group is not None:
            return self.has_warnings(g.formedTree.name2nc[bool_group])
        return super(MpulsViewRenderer, self).has_warnings(nc)


class CoachesPageRenderer(PageRenderer):

    """Respekt-Coaches specific renderer that handles bool-group warnings.
    """

    def get_tree_form_data(self, *args, **kw):
        return CoachesFormData(*args, **kw)


class FormularpageController(MpulsFormularpageController):

    def _get_formpage_renderer(self):
        page_render = CoachesPageRenderer()
        page_render.set_viewrenderer(ViewRenderer)
        return page_render.get_rendered_page
