# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.

import logging
from datetime import datetime, date
from collections import Counter
import json

import paste

from pylons import app_globals as g, response
from pylons.controllers.util import abort

from formed.meta.semantic import SemanticError
from formed.renderer.html import hidden_multiselect_tag

from mpulsweb.lib.translation import _
from mpulsweb.lib.db import db
from mpulsweb.lib.security import checkRole
from mpulsweb.model.logbook import (
    Logbook,
    LogbookEntry,
    build_logbook_entry,
    export_as_xlsx
)
from mpulsweb.model.document import LogbookDocumentAPI, \
     DocumentNotFoundException
from mpulsweb.lib.helpers import url, render_logbook_typeoptions, escape
from mpulsweb.lib.base import DownloadController, c, render, request, session, \
     restrict_to_POST, abort_unless_POST
from mpulsweb.lib.validators import HtmlCleanedString
from mpulsweb.lib.formedforms import FormedForm, Override, Button, \
     file_size_checker, DocumentNewForm, DocumentRenameForm


log = logging.getLogger(__name__)

def get_sort(params):
    sort = params.get('sort_field', 'datum')
    order = params.get('sort_order', 'desc')
    return sort, order


def logbook_art_renderer(renderer, nc, value):
    html_parts = []
    for typ in g.mpuls_config.get('logbook', 'types'):
        if typ["db_field"] == "art":
            html_parts = render_logbook_typeoptions(typ, set(value))
    html_parts.append(hidden_multiselect_tag("art"))
    return "".join(html_parts)

def check_logbook_art(nc, value):
    if not isinstance(value, list):
        value = [value]
    descriptions = g.mpuls_config.get('logbook', 'descriptions')[0]
    checked = []
    for choice in value:
        if choice not in descriptions:
            raise SemanticError("'%s' is not a valid value for '%s'"
                                % (value, nc.getName()))
        checked.append(int(choice))
    return checked


tiny_mce_script = '''\
<script src="%(tinymce_url)s"></script>
<script>
    tinymce.init({
           selector: '#notice',
           menubar: false,
           plugins: 'lists paste',
           contextmenu: false,
           toolbar: 'undo redo | bold italic underline strikethrough | numlist bullist | forecolor backcolor removeformat',
           valid_elements : 'p,strong,span[style=color],em,br,ol,ul,li',
           statusbar: true,
           resize: true,
           elementpath: false,
           branding: false,
           language: 'de',
           browser_spellcheck: true,
           content_css : %(content_css_url)s,
           setup: function(editor) {
               editor.on('change', function(e) { setModification(); });
           }});
</script>
'''

def notice_renderer(renderer, nc):
    item = super(renderer.__class__, renderer).render_text_area(nc)
    item.txt += tiny_mce_script % {
        "tinymce_url": escape(url('/js/tinymce/tinymce.min.js')),
        # For various reasons we don't apply escape to the
        # content_css_url value. For one, the parsing rules for HTML
        # don't make it necessary and even wrong. Browsers apparently
        # treat the contents as raw text and do not expand character
        # entities, even though the official content model for script is
        # CDATA. We do need to make sure that it is a valid javascript
        # string literal, though, hence the use of JSON as a hopefully
        # sufficient approximation (the string's contents cannot be
        # influenced by the user and it's unlikely that the URL contains
        # problematic characters)
        "content_css_url": json.dumps(url('/styles/tinymce_content.css')),
        }
    return item


logbook_document_new_form = DocumentNewForm("logbook", "new_document_action")

class RenameForm(DocumentRenameForm):

    def get_api(self, instance):
        return LogbookDocumentAPI(instance.get("parent_id"))


logbook_document_rename_form = RenameForm("logbook", "rename_document_action")


class LogbookEntryForm(FormedForm):

    def __init__(self, action):
        super(LogbookEntryForm, self).__init__(
            "logbook_entry.xml", "logbook", action, modification_checks=True,
            overrides=[
                Override("art", checker=check_logbook_art,
                         render_choices=logbook_art_renderer),
                Override("notice", validator=HtmlCleanedString(),
                         renderer=notice_renderer),
                Override("file", checker=file_size_checker),
                ],
            buttons=[
                Button(u"Speichern", "save", primary=True),
                Button(u"Speichern und Zurück", "save_and_back"),
                Button(u"Speichern und neuen Eintrag anlegen", "save_and_new"),
                ])

    def create_instance(self):
        instance = super(LogbookEntryForm, self).create_instance()
        instance.set_value("maxupload",
                           g.mpuls_config.get("case", "max-document-size-text"))
        return instance

    def from_object(self, entry):
        instance = self.create_instance()
        instance.set_value("logbook_id", entry.id)
        instance.set_value("date", entry.datum.date())
        instance.set_value("time", entry.datum.time())
        instance.set_value("duration", entry.dauer)
        instance.set_value("bearbeiter", entry.getEditor())
        instance.set_value("art", entry.get_kinds())
        instance.set_value("short_notice", entry.kurz_notiz)
        instance.set_value("notice", entry.notiz)
        return instance


logbook_new_form = LogbookEntryForm("new_action")

logbook_edit_form = LogbookEntryForm("edit_action")


class LogbookController(DownloadController):

    @checkRole(('cm'))
    def index(self, id):
        return self.overview(id)

    def overview(self, id):
        sort, order = get_sort(request.params)
        id = self._checkInt(id)

        ## Load logbook
        logbook = Logbook()
        logbook.loadById(id, sort, order)

        self._ensure_session_case(id)

        c.logbook = logbook
        c.logbook_documents = LogbookDocumentAPI.list_case_logbook_documents(id)
        c.logbook_document_counts = Counter(doc.logbook_id
                                            for doc in c.logbook_documents)

        c.sort_field = sort
        c.sort_order = order

        ## Build evaluations
        c.eval_effort = logbook.getEvaluation()
        c.eval_typ_effort = logbook.getTypEvaluation()

        return render('/logbook/overview.mako')

    @checkRole(('cm'))
    def show(self, id):
        id = self._checkInt(id)
        try:
            entry = LogbookEntry()
            entry.loadById(id)
        except Exception, e:
            c.dialog_title = _(u"Showing of logbook entry failed!")
            c.dialog_text = _(u"An error occured while showing the logbook"
                              u" entry.")
            c.url_ok = url(controller='logbook', action='index',
                           id=session['case'].id)
            log.exception(e)
            return render('/logbook/dialogs/failed_show_logbook_entry.mako')

        self._ensure_session_case(entry.getMasterId())
        c.entry = entry
        c.files = LogbookDocumentAPI(id).list_db()
        return render('/logbook/show.mako')

    def new_js(self):
        response.headers["Content-Type"] = ("application/javascript;"
                                            " charset=utf8")
        return logbook_new_form.form_info_js

    @checkRole('cm')
    def new(self, id):
        id = self._checkInt(id)
        self._ensure_session_case(id)
        if not session["case"].is_editable_state():
            abort(401)

        instance = logbook_new_form.create_instance()
        instance.set_value("case_id", id)
        now = datetime.now()
        instance.set_value("date", now.date())
        instance.set_value("time", now.time())
        instance.set_value("art", [])
        instance.set_value("bearbeiter",
                           session.get('USER_AUTHORIZED').shortenedName())
        instance.set_value("show_file_upload", True)
        return self._render_new_entry_form(instance, {})

    def _render_new_entry_form(self, instance, errors):
        c.rendered_form = logbook_new_form.render(instance, errors)
        c.conditional_form_info_url = url(controller='logbook', action='new_js')
        return render('/logbook/new.mako')

    @checkRole('cm')
    @restrict_to_POST
    def new_action(self):
        instance, errors = logbook_new_form.validate_request(request.params)

        case_id = instance.get("case_id")
        self._ensure_session_case(case_id)
        if not session["case"].is_editable_state():
            abort(401)

        if errors:
            return self._render_new_entry_form(instance, errors)

        log.debug("New Logbook entry from case %s", case_id)
        logbook = Logbook(case_id)
        entry = build_logbook_entry(kinds=instance.get("art"),
                                    date=instance.get("date"),
                                    time=instance.get("time"),
                                    duration=instance.get("duration"),
                                    notice=instance.get("notice"),
                                    short_notice=instance.get("short_notice"),
                                    typ=None)
        logbook.addEntry(entry)
        with db.transaction() as cur:
            logbook.store_cursor(cur, session['USER_AUTHORIZED'].id)
            files = instance.get('file')
            if files:
                api = LogbookDocumentAPI(entry.id)
                api.upload_documents(cur, files)

        c.success_for = _(u"Entry successfully created.")
        c.success_text = _(u"The logbook entry was sucessfully created")
        if request.params.get('save_and_back'):
            return self.overview(case_id)
        elif request.params.get('save_and_new'):
            return self.new(case_id)
        else:
            return self.edit(entry.id)

    def edit_js(self):
        response.headers["Content-Type"] = ("application/javascript;"
                                            " charset=utf8")
        return logbook_edit_form.form_info_js

    def _render_edit_entry_form(self, instance, errors):
        c.rendered_form = logbook_edit_form.render(instance, errors)
        c.conditional_form_info_url = url(controller='logbook',
                                          action='edit_js')
        return render('/logbook/edit.mako')

    @checkRole('cm')
    def edit(self, id):
        id = self._checkInt(id)

        entry = LogbookEntry()
        entry.loadById(id)
        self._ensure_session_case(entry.getMasterId())
        if not entry.is_editable_in_case(session["case"]):
            abort(401)

        c.entry = entry
        instance = logbook_edit_form.from_object(entry)
        return self._render_edit_entry_form(instance, {})

    @checkRole('cm')
    @restrict_to_POST
    def edit_action(self):
        instance, errors = logbook_edit_form.validate_request(request.params)

        logbook_id = instance.get("logbook_id")

        # Abort if editing not allowed. As a side effect it makes sure
        # that session["case"] is set correctly.
        entry = LogbookEntry()
        entry.loadById(logbook_id)
        self._ensure_session_case(entry.getMasterId())
        if not entry.is_editable_in_case(session["case"]):
            abort(401)

        if errors:
            c.entry = entry
            instance.set_value("bearbeiter",
                               session.get('USER_AUTHORIZED').shortenedName())
            return self._render_edit_entry_form(instance, errors)

        entry = build_logbook_entry(kinds=instance.get("art"),
                                    date=instance.get("date"),
                                    time=instance.get("time"),
                                    duration=instance.get("duration"),
                                    notice=instance.get("notice"),
                                    short_notice=instance.get("short_notice"),
                                    typ=None)
        entry.setId(logbook_id)
        entry.storeForUser(session['USER_AUTHORIZED'].id)

        c.success_for = _(u"Changes on logbook successfull!")
        c.success_text = _(u"The changes on the logbook entry were"
                           u" applied successfully.")
        if request.params.get('save_and_back'):
            return self.overview(session['case'].id)
        elif request.params.get('save_and_new'):
            return self.new(session['case'].id)
        else:
            return self.edit(logbook_id)

    @checkRole('cm')
    def delete(self, id, confirmed=False):
        case_id = None
        id = self._checkInt(id)
        try:
            entry = LogbookEntry()
            entry.loadById(id)
            case_id = entry.getMasterId()
            self._ensure_session_case(case_id)
        except Exception, e:
            log.exception(e)
            c.dialog_title = _(u"Checking of logbook entry delete permission"
                             u" failed!")
            c.dialog_text = _(u"An error occured while checking delete"
                              u" permission for the logbook entry.")
            c.url_ok = url(controller='logbook', action='index',
                           id=(case_id if case_id is not None
                               else session["case"].id))
            return render('/logbook/dialogs/failed_delete_logbook_entry.mako')

        if not entry.is_editable_in_case(session["case"]):
            # The case is either not editable or the entry to be deleted
            # is an automatic entry. In both cases, deletion is not
            # permitted
            abort(401)

        confirmed = self._checkBool(confirmed)
        if confirmed == 1:
            abort_unless_POST()
            try:
                Logbook().deleteEntryById(int(id))
            except Exception, e:
                log.exception(e)
                c.dialog_title = _(u"Deletion of logbook entry failed!")
                c.dialog_text = _(u"An error occured while deleting the"
                                  u" logbook entry.")
                c.url_ok = url(controller='logbook', action='index', id=case_id)
                return render('/logbook/dialogs/failed_delete_logbook_entry.mako')
            return self._delete_succeded(case_id)
        else:
            c.confirm_for = _(u"Delete logbook entry?")
            c.question = _(u"Do you want to delete this logbook entry?\n"
                           u"The data will be irrevocably lost.")
            c.url_yes = url(controller='logbook', action='delete', id=id,
                            confirmed=1)
            c.url_no = url(controller='logbook', action='index', id=case_id)
            return render('/logbook/dialogs/confirm_delete_logbook_entry.mako')

    def _delete_succeded(self, case_id):
        c.success_for = _(u"Logbook entry successfully deleted!")
        c.success_text = _(u"Logbook entry was successfully deleted!")
        return self.index(case_id)

    def _render_new_logbook_document_form(self, instance, errors):
        c.rendered_form = logbook_document_new_form.render(instance, errors)
        return render('/logbook/newdocument.mako')

    @checkRole('cm')
    def newDocument(self, id):
        id = self._checkInt(id)
        c.entry = LogbookEntry()
        c.entry.loadById(id)
        self._ensure_session_case(c.entry.getMasterId())
        instance = logbook_document_new_form.create_instance()
        instance.set_value("id", id)
        return self._render_new_logbook_document_form(instance, {})

    @checkRole('cm')
    @restrict_to_POST
    def new_document_action(self):
        instance, errors = logbook_document_new_form.validate_request(
                                                                 request.params)

        logbook_id = instance.get("id")
        if errors:
            c.entry = LogbookEntry()
            c.entry.loadById(logbook_id)
            return self._render_new_logbook_document_form(instance, errors)

        c.entry = LogbookEntry()
        c.entry.loadById(logbook_id)
        self._ensure_session_case(c.entry.getMasterId())
        if not session["case"].is_editable_state():
            abort(401)

        api = LogbookDocumentAPI(logbook_id)
        api.upload_documents_transaction(instance.get("file"))

        c.success_text = _(u"Attachment added!")
        return self.show(logbook_id)


    @checkRole(('admin', 'cm'))
    def showDocument(self, id):
        id = self._checkInt(id)
        api = LogbookDocumentAPI(None)
        try:
            document = api.load_db(id, load_contents_immediately=True)
        except DocumentNotFoundException:
            # the document could not be found, so we should send a 404
            # response to the client.
            abort(404)

        data = paste.fileapp.DataApp(document.data,
                                     content_type=document.mime,
                                     cache_control='must-revalidate',
                                     pragma='public')
        return data(request.environ, self.start_response)


    def _load_editable_document(self, doc_id):
        doc_id = self._checkInt(doc_id)
        api = LogbookDocumentAPI(None)
        try:
            doc = api.load_db(doc_id, load_contents_immediately=False)
        except DocumentNotFoundException:
            abort(404)

        # Abort if editing not allowed. As a side effect it makes sure
        # that session["case"] is set correctly.
        entry = LogbookEntry()
        entry.loadById(doc.logbook_id)
        self._ensure_session_case(doc.case)
        if not entry.is_editable_in_case(session["case"]):
            abort(401)

        c.entry = entry
        return doc

    @checkRole('cm')
    def deleteDocument(self, id, confirmed):
        doc = self._load_editable_document(id)

        confirmed = self._checkBool(confirmed)
        if confirmed == 1:
            abort_unless_POST()
            LogbookDocumentAPI(doc.logbook_id).delete_transaction(doc.id)
            c.url_ok = url(controller="logbook", action="show",
                           id=doc.logbook_id)
            c.success_text = _(u"Attachment successfully deleted!")
            return self.show(doc.logbook_id)
        else:
            c.context = "../main.mako"
            c.confirm_for = _(u"Delete attachment?")
            c.question = _(u"Do you want to irrevocably delete the attachment"
                           u" '%s'?") % doc.name
            c.url_yes = url(controller='logbook', action='deleteDocument',
                            id=id, confirmed=1)
            c.url_no = url(controller="logbook", action="show",
                           id=doc.logbook_id)
            return render('/logbook/dialogs/'
                          'confirm_delete_document.mako')

    def _render_rename_form(self, instance, errors):
        c.rendered_form = logbook_document_rename_form.render(instance, errors)
        return render('/logbook/renamedocument.mako')

    @checkRole('cm')
    def renameDocument(self, id):
        doc = self._load_editable_document(id)
        instance = logbook_document_rename_form.from_object(doc)
        instance.set_value("parent_id", doc.logbook_id)
        c.logbook_id = doc.logbook_id
        return self._render_rename_form(instance, {})

    @checkRole('cm')
    @restrict_to_POST
    def rename_document_action(self):
        instance, errors = logbook_document_rename_form.validate_request(
                                                                request.params)

        if errors:
            return self._render_rename_form(instance, errors)

        id = instance.get("id")

        doc = self._load_editable_document(id)

        api = LogbookDocumentAPI(doc.logbook_id)
        api.rename_transaction(doc.id, instance.get("name"))

        c.success_text = _(u"Attachment successfully renamed.")
        return self.show(doc.logbook_id)

    @checkRole(('cm'))
    def printout(self, id):
        id = self._checkInt(id)
        c.print_version = 1

        # Load logbook
        logbook = Logbook()
        logbook.loadById(id)
        c.logbook = logbook

        # Load case
        case = self._loadCase(logbook.getMasterId(), read_only=True)
        c.case = case

        return render('/logbook/overview_print.mako')

    @checkRole(('cm',))
    def export(self, id):
        id = self._checkInt(id)

        # Load logbook for id
        logbook = Logbook()
        logbook.loadById(id)

        case = self._loadCase(logbook.getMasterId(), read_only=True)

        # Work happens in export_as_xls
        return self._download(export_as_xlsx(logbook, case),
                              ("application/vnd.openxmlformats-"
                               "officedocument.spreadsheetml.sheet"),
                              "tagebuch-export", extension="xlsx")
