# Copyright 2007, 2008, 2010 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.

"""The base Controller API

Provides the BaseController class for subclassing, and other objects
utilized by Controllers.
"""

import hashlib
import logging
import locale

import formencode

import paste

import pylons.decorators.rest

from formed.meta.data import PageNode
from formed.renderer.html import repeat_id

import mpulsweb.lib.db as db
import mpulsweb.lib.helpers as h

from pylons import tmpl_context as c, config, app_globals as g, \
    request, response, url
from pylons.controllers import WSGIController
from pylons.controllers.util import redirect, abort
from pylons.templating import cached_template, pylons_globals

from webhelpers.html import literal

from webob.exc import HTTPUnauthorized, HTTPNotFound

from formed.instance.tree import IDENTIFIER_SEPARATOR

from mpulsweb.config.importer import import_overridable_module
from mpulsweb.lib.translation import set_lang, _, ungettext
from mpulsweb.lib.app_globals import session
from mpulsweb.lib.logfilter import DBNameContextFilter
from mpulsweb.lib.security import userIdentity, get_db_parameters, \
    current_csrf_token, check_csrf_token
from mpulsweb.lib.exceptions import LoadCaseNotExistsError, \
    MpulsMissingItemException
from mpulsweb.lib.customoverview import Column, Overview
import mpulsweb.lib.helper.filters as F


# These names are exported from this module largely for historical
# reasons.  Ideally the module should probably only export
# BaseController and render which are customarily imported from this
# module in Pylons.
__all__ = ["_", "BaseController", "c", "g", "h", "redirect", "render",
           "render_htmlfill", "request", "response", "session"]


COOKIE_NOT_FOUND = ("Cookie konnte nicht gefunden werden."
                    " Haben Sie Cookies in Ihrem Browser aktiviert?")
IDENTITY_CHANGED = "Ihre Identitaet hat sich geaendert."

log = logging.getLogger(__name__)


# Decorator for actions that handle POST request.
# With active CSRF-protection if forms the requests resulting from those
# forms need to be restricted to POST requests. If GET requests were
# allowed, too, the CSRF-protection wouldn't work, because the
# CSRF-token is only used together with POST.
restrict_to_POST = pylons.decorators.rest.restrict("POST")


def abort_unless_POST():
    """Abort the current request if it is not a POST request.
    Intended use is in actions that handle both GET and POST requests
    where GET requests are read-only and POST-requests modify data or
    have other side-effects and use CSRF-protection.

    The HTTP response is 404 because that is the most generic way to
    deny a request without having to give reasons. A 405 response as
    used by restrict_to_POST could also be appropriate, but since this
    function is to be used by methods that handle other HTTP methods
    besides POST would not always be appropriate because it would, often
    incorrectly, imply that the URI of the request can only be accessed
    with POST.
    """
    if request.method != "POST":
        abort(404)


def auth_cookie_name():
    """Return the name of authorization cookie.
    The name is taken from the config setting mpuls.app.auth_cookie_name
    and defaults to 'mpuls_auth'.
    """
    return config.get("mpuls.app.auth_cookie_name", "mpuls_auth")


# Added own render_mako function which addes custom translation functions
def render_mako(template_name, extra_vars=None, cache_key=None,
                cache_type=None, cache_expire=None):
    """Render a template with Mako

    Accepts the cache options ``cache_key``, ``cache_type``, and
    ``cache_expire``.

    """
    # Create a render callable for the cache function
    def render_template():
        # Pull in extra vars if needed
        globs = extra_vars or {}

        # Second, get the globals
        globs.update(pylons_globals())

        # Thirs overwritw '_'
        globs.update({'_': _, 'ungettext': ungettext})
        globs["session"] = session

        globs["g"] = g

        # Grab a template reference
        template = globs['app_globals'].mako_lookup.get_template(template_name)

        return literal(template.render_unicode(**globs))

    return cached_template(template_name, render_template, cache_key=cache_key,
                           cache_type=cache_type, cache_expire=cache_expire)
render = render_mako


def render_htmlfill(form, defaults=None, *args, **kwargs):
    """mpuls specific frontend for formencode.htmlfill.render.
    This function passes all positional and keyword arguments to
    formencode.htmlfill.render, with one modification:
    The CSRF-token value is is added to the dictionary given as the
    value of the 'defaults' argument.
    """
    csrf_defaults = {"csrftoken": current_csrf_token()}
    if defaults is not None:
        defaults = defaults.copy()
    else:
        defaults = {}
    defaults.update(csrf_defaults)
    return formencode.htmlfill.render(form, defaults, *args, **kwargs)


class CaseMixin(object):

    def _loadCase(self, id, uncommitted=None, read_only=False):
        """Load and return the case object for the case with the given id.

        The parameter uncommitted, if given and not None, should be a
        dictionary usable as the uncommitted parameter of the loadById
        method of the case factory.

        If the parameter read_only is false (the default), the case must
        be editable by the current user as determined by the case's
        user_may_edit method. If the case is not editable in this sense,
        the method calls the showError method. Also, showError is called
        if the case is not readable by the current user as determined by
        the case's user_may_read method.

        Note that the default value of the read_only parameter means
        that by default the _loadCase assumes that cases are loaded in
        order to modify them and will raise an exception if the user is
        not allowed to do that. The default value was chosen to err on
        the side of caution. From a security point of view it's better
        to deny an action that should have been allowed, than to allow
        an action that ought to have been denied.
        """
        try:
            case = g.case_factory.loadById(id, uncommitted=uncommitted)
        except LoadCaseNotExistsError as e:
            log.error("Could not load case %r from %r, sending 404",
                      e.case_id, e.dbname)
            self.showError()

        user = session["USER_AUTHORIZED"]
        if not case.user_may_read(user):
            self.showError()
        if not read_only and not case.user_may_edit(user):
            self.showError()

        session_case = case.getSessionCase()
        if "case" in session:
            old_caseid = session.get('case').id
        else:
            old_caseid = None

        # New session code
        session['case'] = session_case

        # Only reset navigation if case is switched
        if str(old_caseid) != str(id) or \
                'navigation.tree' not in session \
                or 'navigation.tree.recreate' in session:
            if 'navigation.tree.recreate' in session:
                del session['navigation.tree.recreate']
            open_folders = []
            navigation = g.get_navigation(id)
            default_render_mode = 'ro'
            if user.getPermissionRole() == "cm":
                default_render_mode = g.mpuls_config.get_default_render_mode()
            session['render_mode'] = default_render_mode
            session['navigation.tree'] = navigation
            session['navigation.openfolders'] = open_folders
            session['navigation.selectedpage'] = None
            # Also reset uncommited fields, because a new case is loaded.
            session['uncommited_fields'] = {}
        session.save()
        return case

    def _load_case_for_node_id(self, node_id, read_only=False):
        """Load the case object for a given node ID.

        The node ID can be e.g. a page ID or other node ID that is a
        string that ends with the case ID separated by a colon.
        """
        case_id = self._checkInt(node_id.rpartition(IDENTIFIER_SEPARATOR)[-1])
        return self._loadCase(case_id)

    def _ensure_session_case(self, case_id):
        """Ensure that the session-case refers to the given case.

        If the session case (stored in the session dictionary under
        "case") is not set or does not refer to the case indicated by
        the case ID given, use self._loadCase to set it. This may cause
        database round-trips in order to load the case object.

        This method is useful where information is processed that
        belongs to a particular case but where the case object itself is
        not actually needed. For example, the logbook is related to a
        case and when rendering the templates the session-case is needed
        but for the logbook itself the case object is not needed.
        """
        session_case = session.get('case')
        if session_case is None or session_case.id != case_id:
            self._loadCase(case_id)

    def _load_current_case(self, read_only=False):
        """Load and return the case object for the session's current case.

        The session's current case is the one stored in the session case
        in session['case'].

        The case is loaded with the _loadCase method. The parameter
        read_only is simply passed through to _loadCase. See the
        documentation of _loadCase for the semantics of read_only and
        which exception might be raised.
        """
        return self._loadCase(session['case'].id,
                              session.get('uncommited_fields'),
                              read_only=read_only)

    def _handle_automatic_phase_change(self, case, default_url):
        """Change Phase automatically depending on the state of the case"""
        fullautomatic_title = _("Phase automatically changed")
        halfautomatic_title = _("Change phase?")

        fullautomatic_text = _(
            "The phase of the case was changed automatically"
            " after checking the filled out required fields."
            " The phase changed from '%s' to '%s'."
            " Please click on 'Ok', to proceed editing"
            " the case.")
        halfautomatic_text = _(
            "After checking the filled out required fields,"
            " it seems that a phase change can be done from"
            " phase '%s' to '%s'."
            " Please click on 'Yes' to perform the change."
            " Click on 'No' to perform the change later"
            " manually on the phase overview page.")

        sm = g.case_factory.get_state_machine()
        changemode = sm.get_transition_mode()
        if changemode != sm.TRANSITION_MODE_MANUAL:
            current_phase = sm.get_current_state(case)
            reachable_phase = sm.get_next_state(case)
            log.debug("current phase: %s" % current_phase)
            log.debug("max phase? %s" % reachable_phase)
            if current_phase != reachable_phase:
                cpd = current_phase.get_display_name()
                rpd = reachable_phase.get_display_name()
                if changemode == sm.TRANSITION_MODE_FULLAUTO:
                    log.debug("fullautomatic phase change!")
                    case.setPhase(reachable_phase.get_id())
                    # build dialog
                    c.dialog_title = fullautomatic_title
                    c.dialog_text = literal((fullautomatic_text % (cpd, rpd)))
                    c.url_ok = default_url
                    return render(
                        '/phase/dialogs/success_fullautomatic_set_phase.mako')
                else:
                    log.debug("halfautomatic phase change!")
                    # build dialog
                    c.dialog_title = halfautomatic_title
                    c.dialog_text = literal((halfautomatic_text % (cpd, rpd)))
                    c.url_ok = url(controller="phase", action="setPhase",
                                   id=case.id,
                                   value=reachable_phase.get_id())
                    c.url_no = default_url
                    return render(
                        '/phase/dialogs/success_halfautomatic_set_phase.mako')
        return None


class RGMixin(object):

    def _add_rg(self, case, rg_id, anchor=None):
        """Create a new repeat group instance and return its URL"""
        try:
            rg_instance = case.create_repeat_group_instance(rg_id)
        except MpulsMissingItemException:
            abort(404, _('repeat group container not found.'))

        # Build new Navigation
        navigation = g.get_navigation(session['case'].id,
                                      session['navigation.openfolders'])
        session['navigation.tree'] = navigation
        session.save()

        # Determine which page to redirect to. When the repeat group is
        # embedded in a form page, we stay on the selected page
        # (navigation.selectedpage). When it's a standalone repeat
        # group, as indicated by the meta node having a page as a child,
        # we redirect to the page of the new instance.
        page_id = session['navigation.selectedpage']
        first_child = rg_instance.getMeta().getChildren()[0]
        if isinstance(first_child, PageNode):
            page_name = first_child.getName()
            page_id = page_name + ":" + repeat_id(rg_instance.getIdentifier())

        anchor_args = {}
        if anchor is not None:
            anchor_args["anchor"] = anchor
        return h.url(controller='formularpage', action='select',
                     id=page_id, **anchor_args)

    def _add_rg_and_redirect(self, case, rg_id, anchor=None):
        """Create a new repeat group instance and return its URL"""
        rg_url = self._add_rg(case, rg_id, anchor=anchor)
        redirect(rg_url)


class SessionMixin(object):

    def _register_session(self, environ):
        mpuls_session_module = import_overridable_module("lib.session")
        environ['paste.registry'].register(session,
                                           mpuls_session_module.MPulsSession())


class CsrfMixin(object):

    def _check_csrf_token(self):
        if g.mpuls_config.is_enabled('security', 'require-csrf-tokens'):
            if not check_csrf_token(request):
                log.error("CSRF-token check failed. Sending 401 Unauthorized")
                raise HTTPUnauthorized()


class SessionAuthMixin(object):

    def _check_session_auth(self):
        if "AUTH" not in session:
            return

        storedHash = session['AUTH']
        try:
            shared = request.cookies[auth_cookie_name()]
        except KeyError:
            log.error("Cannot find %s cookie -> HTTPUnauthorized",
                      auth_cookie_name())
            session.delete()  # session should be removed on errors
            raise HTTPUnauthorized(detail=COOKIE_NOT_FOUND)

        if storedHash != hashlib.md5(shared + userIdentity()).digest():
            log.error("Stored hash doesn't match new hash "
                      "-> HTTPUnauthorized")
            session.delete()  # session should be removed on errors
            raise HTTPUnauthorized(detail=IDENTITY_CHANGED)


class UserMixin(object):

    def _check_user_authorized(self):
        # Check if the user is already authorized.
        # If not then redirect the user to the login page
        if 'USER_AUTHORIZED' not in session:
            # p = str(request.path_info)
            # if p not in ('/auth/login', '/auth/loginAction', '/auth/logout'):
            log.debug("User not found in session. Deleting session and "
                      "redirect user to login page")
            session.delete()  # session must be deleted on errors
            redirect(config.get("mpuls.auth.login_url")
                     or h.url(controller='auth', action='login'))

    def _set_db_for_user(self):
        self.user = session['USER_AUTHORIZED']
        DBNameContextFilter.set_thread_dbname(self.user.agency)

        db_parameters = get_db_parameters(self.user.login, self.user.password,
                                          config.get('mpuls.db.host'),
                                          config.get('mpuls.db.port'),
                                          config.get('mpuls.db.database'),
                                          config.get('mpuls.db.user'))
        log.debug("DB parameters: %r", db_parameters)
        DBNameContextFilter.set_thread_dbname(db_parameters["database"])
        self.db = db.DB(**db_parameters)
        db.enter(self.db)

    def _check_user(self):
        # Is the user activated?
        log.debug('User accound activated: %s' % self.user.isActivated())
        log.debug('User needs new pass: %s' % self.user.needsNewPass())
        if (not self.user.isActivated() and
                request.path_info not in ['/auth/accountDeactivated']):
            redirect(url('/auth/accountDeactivated'))

        # Do the user needs to change his passwort?
        elif (self.user.needsNewPass() and
                request.path_info not in
                ['/usersettings/changePasswordOnLogin',
                 '/usersettings/changePasswordOnLoginAction',
                 '/auth/accountDeactivated']):
            redirect(url('/usersettings/changePasswordOnLogin'))


class DownloadMixin(object):

    def _download(self, data, mime_type, name, extension, attachment=True,
                  **extra):
        if attachment:
            disposition = 'attachment; filename="%s.%s"' % (name, extension)
        else:
            disposition = 'filename="%s.%s"' % (name, extension)
        data = paste.fileapp.DataApp(data,
                                     content_type=mime_type,
                                     content_disposition=disposition,
                                     cache_control='must-revalidate',
                                     pragma='public',
                                     **extra)
        return data(request.environ, self.start_response)


class CaseOverviewMixin(object):

    def _get_default_case_overview_columns(self):
        columns = g.case.get_overview_fields_list()
        columns.extend([
            Column("bearbeiter_id", _('Editor'), "table_name",
                   ["admin"],
                   lambda case: "%s, %s" % (F.NA(case.getEditor().last_name),
                                            F.NA(case.getEditor().first_name[:1]))),
            Column("zugriff", _('Updated'), "table_name", ["cm"],
                   lambda case: h.format_date(case.state.getAccessTime())),
            Column("id", "ID", "table_width_mid_s", ["cm", "admin"],
                   lambda case: F.NA(case.id)),
            ])
        return columns


class BaseCsrfController(SessionMixin, CsrfMixin, WSGIController):

    def __before__(self, environ):
        self._register_session(environ)
        self._check_csrf_token()


class BaseController(CaseMixin, SessionMixin, SessionAuthMixin, CsrfMixin,
                     UserMixin, RGMixin, WSGIController):

    def __init__(self):
        self.db = None

    def __before__(self, environ):
        self._register_session(environ)
        self._check_csrf_token()
        self._check_session_auth()

        self._check_user_authorized()
        self._set_db_for_user()
        self._check_user()

        # Set language for translation based on browser setting. Add a
        # default language based on the server's locale settings.
        deflang = locale.getlocale()[0]
        all_languages = request.languages + [deflang]
        set_lang(all_languages)
        log.debug('Language set to %s', all_languages)

        # Allow for application specific checks. Done last in __before__
        # so that the standard checks have been made and the databse
        # connection has been set up.
        g.before_hook()

    def __call__(self, environ, start_response):
        """Invoke the Controller"""
        # WSGIController.__call__ dispatches to the Controller method
        # the request is routed to. This routing information is
        # available in environ['pylons.routes_dict']
        try:
            return WSGIController.__call__(self, environ, start_response)
        finally:
            _db = self.db
            if _db:
                self.db = None
                db.leave(_db)
                _db.closeConnections()
            DBNameContextFilter.set_thread_dbname(None)

    def _checkInt(self, i):
        try:
            return int(i)
        except (TypeError, ValueError):
            self.showError()

    def _checkBool(self, bool):
        confirmed = self._checkInt(bool)
        if confirmed not in [0, 1]:
            self.showError()
        return confirmed == 1

    def showError(self, err=None):
        raise HTTPNotFound()


class DownloadController(DownloadMixin, BaseController):
    pass
