# -*- coding: utf-8 -*-
import logging

from formed.meta.data import ChoiceNode, DateLeaf, IntLeaf, NumericLeaf, \
     TextLeaf, TextAreaLeaf, BoolLeaf, RadioNode, VirtualLeaf, ReferenceNode
from mpulsweb.lib.base import g
from mpulsweb.lib.translation import _
from mpulsweb.model.state import PhaseStateMachine
from mpulsweb.model.user import UserListObject


log = logging.getLogger(__name__)

def dump_normal_fields(filter=(), formed_tree=None):
    if formed_tree is None:
        formed_tree = g.formedTree
    fields = []
    for page, nodes in formed_tree.page_fields.items():
        page_node = formed_tree.findByName(page)
        if page_node.isInvisible():
            continue
        fields += [dict(description=nc.get_title(), name=nc.getName())
                   for nc in nodes
                   if not isinstance(nc, VirtualLeaf)
                   if not isinstance(nc, ReferenceNode)
                   if nc.getName() not in filter]

    return fields


def dump_special_fields(filter=()):
    extras = [STATIC_EXTRAS, DYNAMIC_EXTRAS]

    if g.mpuls_config.is_enabled("case-module", "logbook"):
        extras.append(LOGBOOK_EXTRAS)

    fields = [dict(description=value.get("description"), name=key)
              for field_dict in extras
              for key, value in field_dict.items()
              if key not in filter]
    return fields


def dump_all_fields(filter=()):
    """Returns all searchable fields as a list of dictionaries.
    Each dictionary contains the title, name and table of the field,
    stored with the keys "description", "name" and "table" respectively.
    "Searchable" are visible fields in the default formed tree (g.formedTree).
    And some "special" fields like editor or phase.
    ReferenceNodes are also not searchable.

    All fields named in the iterable filter are ignored.

    Virtual Fields are ignored.
    """
    return dump_normal_fields(filter) + dump_special_fields(filter)

def description_to_name(description):
    fields = dump_all_fields()
    for field in fields:
        if field.get("description") == description:
            return field.get("name")

def nc_type_name(nc):
    if isinstance(nc, DateLeaf):
        return "date"
    elif isinstance(nc, IntLeaf):
        return "int"
    elif isinstance(nc, NumericLeaf):
        return "float"
    elif isinstance(nc, (TextLeaf, TextAreaLeaf)):
        return "text"
    elif isinstance(nc, ChoiceNode):
        return "choice"
    elif isinstance(nc, RadioNode):
        return "radio"
    elif isinstance(nc, BoolLeaf):
        return "bool"
    else:
        raise ValueError("Unexpected node component type: %r" % (nc,))


class SearchFieldDescription(object):

    """Description of a field in the search form.

    Attributes:
        type (str): field type as determined by e.g. nc_type_name
        description (str): Label of the field in the form
        extras: type specific extra information. For 'choice' fields,
            this is a list of key/value pairs with the options.
        error (optional formencode.Invalid): optional validation error

    For backwards compatibility, this class provides an interface
    resembling that of a dictionary because the class replaces an actual
    dictionary.
    """

    def __init__(self, type, description, extras=None, error=None):
        self.type = type
        self.description = description
        self.extras = extras
        self.error = error

    def _lookup(self, key, *args):
        return getattr(self, key, *args)

    def __setitem__(self, key, value):
        return setattr(self, key, value)

    def __getitem__(self, key):
        return self._lookup(key)

    def get(self, key, default=None):
        return self._lookup(key, default)

    def copy(self):
        return self.__class__(self.type, self.description, self.extras,
                              self.error)

    def with_error(self, error):
        copy = self.copy()
        copy.error = error
        return copy



STATIC_EXTRAS = {
    "search_str": SearchFieldDescription(u"text",
                                         u"Name/Vorname"),
    "responsibilities": SearchFieldDescription(u"choice",
                                               u"Bearbeitung",
                                               extras=[[1, u"Eigene Fälle"],
                                                       [2, u"Vertretungsfälle"],
                                                       [3, u"Übertragene Fälle"]]),
    "status": SearchFieldDescription(u"choice",
                                     u"Status",
                                     extras=[[1, u"Bearbeitbare Fallakten"],
                                             [3, u"Schwebend gelöschte"],
                                             [4, u"Schwebend anonymisiert"],
                                             [5, u"Anonymisiert"]]),
    "id": SearchFieldDescription(u"int",
                                 u"ID"),
    "updated": SearchFieldDescription(u"date",
                                      u"Aktualisiert"),
}

# Fields where the extra info is dynamically generated
DYNAMIC_EXTRAS = {
    "phase": SearchFieldDescription("choice",
                                    u"Phase"),
    "bearbeiter_id": SearchFieldDescription("choice",
                                            u"Verantwortlichkeit"),
}

LOGBOOK_EXTRAS = {
    "art": SearchFieldDescription("choice",
                                  u"Tagebucheintrag - Art"),
# "dauer" is commented out as dauer is of type interval in the
# database. We don't support searching in intervals currently.
#
#    "dauer": {"type": "int",
#              "description": u"Tagebucheintrag - Dauer",
#              "extras": None},
    "kurz_notiz":
    SearchFieldDescription("text",
                           u"Tagebucheintrag - Nennung Sonstiges"),
    "notiz": SearchFieldDescription("text",
                                    u"Tagebucheintrag - Notiz"),
    "datum": SearchFieldDescription("date",
                                    u"Tagebucheintrag - Datum"),
}

def is_special_field(name):
    return name in STATIC_EXTRAS.keys() or name in DYNAMIC_EXTRAS.keys() or \
            (g.mpuls_config.is_enabled("case-module", "logbook") and \
             name in LOGBOOK_EXTRAS)

def extra_info(formedtree, name):
    if name in STATIC_EXTRAS:
        return STATIC_EXTRAS[name]

    if name == "art":
        log_art_extras = g.mpuls_config.get("logbook", "descriptions")
        return SearchFieldDescription("choice",
                                      u"Tagebucheintrag - Art",
                                      extras=[[int(key), value]
                                              for key, value
                                              in log_art_extras[0].items()])
    if name in LOGBOOK_EXTRAS:
        # Other logbook extras are static.
        return LOGBOOK_EXTRAS[name]

    if name == "phase":
        # Build the current phases
        state_machine = PhaseStateMachine()
        return SearchFieldDescription(
            "choice",
            u"Phase",
            extras=[(p.id, p.getDescription())
                    for phase in state_machine.phaselist
                    for p in (phase.start, phase.end)])

    if name == "bearbeiter_id":
        ulo = UserListObject()
        filter1 = [u.id for u in ulo.getAdminList()]
        userlist = ulo.getUserList(filter1)
        return SearchFieldDescription(
            "choice", _("Responsibility"),
            extras=[[u.id, u.last_name + ", " + u.first_name]
                    for u in userlist])

    nc = formedtree.name2nc[name]
    type_name = nc_type_name(nc)
    if type_name == "choice" or type_name == "radio":
        extras = nc.get_options()
    elif type_name in ("int", "float"):
        extras = (nc.getMinValue(), nc.getMaxValue())
    else:
        extras = None
    return SearchFieldDescription(type_name, nc.get_title(),
                                  extras=extras)
