# -*- coding: utf-8 -*-
# Copyright 2014 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of MPULS (CoMPUter-based case fiLeS).
#
# MPULS is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# MPULS is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with MPULS. If not, see <http://www.gnu.org/licenses/>.

"""
Classes for a high-level representation of search options.

For background information see MEP-31.
"""
# for the to_json method of Expr
import json
# for to_searchtree
import sys
import datetime

from pylons import app_globals as g

import formed.meta.data

import mpulsweb.lib.sqlexpr as E
from mpulsweb.lib.base import session
from mpulsweb.lib.security import hasRole
from mpulsweb.config.importer import import_overridable_module
from mpulsweb.lib.helper.searchfield import LOGBOOK_EXTRAS

def json_serializer(obj):
    """JSON Serializer that serializes datetime.date objects, too."""
    if isinstance(obj, datetime.date):
        return {"cls": "datetime.date",
                "year": obj.year, "month": obj.month, "day": obj.day}
    return obj


def json_deserializer(obj):
    """JSON deserializer that decodes date objects encoded by json_serializer
    """
    if obj.get("cls") == "datetime.date":
        return datetime.date(obj["year"], obj["month"], obj["day"])
    return obj


def field_to_rg(field):
    """ Get the name of the repeat group a field belongs to. """
    if field in LOGBOOK_EXTRAS:
        # The database relation of logbook entries to the master
        # table is just like a repeat group"
        return "ka_logbuch"
    try:
        s_path = g.structureTree.structure_path(field)
    except KeyError:
        # When a field is not part of the formed tree
        # it is special and also does not have a repeatgroup.
        return None
    if s_path and len(s_path) > 1:
        return s_path[-2].getName()
    return None


def qualified_name(field_name):
    repeat_group = field_to_rg(field_name)
    if repeat_group:
        prefix = repeat_group + "_tbl_view."
    else:
        prefix = "m."
    return prefix + field_name


class Expr(object):

    """Base-class for highlevel mpuls search expressions"""

    def __init__(self):
        pass

    def sql_expr(self):
        raise NotImplementedError

    def to_simple_format(self):
        raise NotImplementedError

    def to_json(self):
        return json.dumps(self.to_simple_format(), default=json_serializer)

    def retrieve_fields(self):
        """Return the fields that the search should retrieve for every match.
        This method should return an iterable that gives the expressions
        to use in the search's SELECT clause so that the value that
        matches the user's search criteria as far as self is concerned
        is returned in the search result and is immediately available
        when showing the search result to avoid a separate return trip
        to the database.

        Note that the iterable must be empty if the parameters of self
        indicate that the user does not want to actually use the
        criterion represented by self. E.g. for Field, a value of ''
        means that the user did not actually type anything into the
        input element for this field in the search form and thus is not
        interested in seeing the field in the results.

        The default implementation returns an empty tuple.
        """
        return ()

    def form_key(self):
        raise NotImplementedError("form_key not implemented for %r" % self)


class _TRUE(Expr):

    """Search condition that's always true.

    Do not instantiate this class.
    Use the module level constant TRUE instead.
    """

    def __repr__(self):
        return "TRUE"

    def to_simple_format(self):
        return {
            'cls': 'TRUE',
            'attrs': {}
        }

    def sql_expr(self):
        return E.TRUE


TRUE = _TRUE()


def field_sql_expr(qualified_name, op, value):
    """Construct SqlExpr for a field comparison.

    This implements the main logic used by the Field class when creating
    the SqlExpr. This is a separate function to make it easier to reuse
    it in special Expr sub-classes that are similar to Field, but
    different enough not to be derived from Field.
    """
    # NOTE: the logic here has to be consistent with the one in
    # Field.retrieve_fields()!
    if issubclass(op, E.Unary):
        return op(E.Field(qualified_name))

    if value is None or value == "":
        return E.TRUE
    return op(E.Field(qualified_name), E.Const(value))


class Field(Expr):

    """Search for cases where a given formed field has a specific value."""

    def __init__(self, fieldname, value, op=E.EQ):
        self.fieldname = fieldname
        self.op = op
        self.value = value

    def __repr__(self):
        return "Field(%r, %r, %r)" % (self.fieldname, self.value, self.op)

    def to_simple_format(self):
        return {
            'cls': self.__class__.__name__,
            'attrs': {"fieldname": self.fieldname,
                      "op": self.op.__name__,
                      "value": self.value,
                      },
        }

    def sql_expr(self):
        return field_sql_expr(qualified_name(self.fieldname), self.op,
                              self.value)

    def retrieve_fields(self):
        # NOTE: the logic for the here has to be consistent with the one
        # in field_sql_expr!
        if (issubclass(self.op, E.Unary)
            or (self.value is not None and self.value != "")):
            yield qualified_name(self.fieldname)

    def form_key(self):
        return self.fieldname


class DayTimestamp(Field):

    """Search for timestamp with the precision of "on that day".
    There is no need / interface to search for a precise
    timestamp at this moment.
    This class casts the timestamp to a date for searching.
    """

    def sql_expr(self):
        return field_sql_expr(qualified_name(self.fieldname) + "::date",
                              self.op, self.value)


class Bool(Field):

    """Search a given bool field. A value of -1 also includes cases
       where the value is not set."""

    def __repr__(self):
        return "Bool(%r, %r)" % (self.fieldname, self.value)

    def sql_expr(self):
        if self.value == None:
            return E.TRUE

        if self.value == 1:
            return E.EQ(E.Field(qualified_name(self.fieldname)),
                                      E.Const(self.value))

        return E.NAryOr([E.EQ(E.Field(qualified_name(self.fieldname)),
                                      E.Const(self.value)),
                         E.IS_NULL(E.Field(qualified_name(self.fieldname)))])


def between_sql_expr(qualified_name, low, high):
    """Construct SqlExpr to test whether a field's values lies in a range.
    This implements the main logic used by the Between class when
    creating the SqlExpr. It is a separate function to make it easier to
    reuse it in special Expr sub-classes that are similar to Between,
    but different enough not to be derived from Field.
    """
    subexprs = []
    field = E.Field(qualified_name)
    if low is not None:
        subexprs.append(E.GE(field, E.Const(low)))
    if high is not None:
        subexprs.append(E.LE(field, E.Const(high)))
    if subexprs:
        return E.NAryAnd(subexprs)

    return E.TRUE


class Between(Expr):

    """Search condition testing that a formed field's values lies in a range.
    """

    def __init__(self, fieldname, low, high):
        self.fieldname = fieldname
        self.low = low
        self.high = high

    def __repr__(self):
        return "Between(%r, %r, %r)" % (self.fieldname, self.low, self.high)

    def sql_expr(self):
        return between_sql_expr(qualified_name(self.fieldname), self.low,
                                self.high)

    def to_simple_format(self):
        return {
            'cls': self.__class__.__name__,
            'attrs': {"fieldname": self.fieldname,
                      "low": self.low,
                      "high": self.high,
                      },
        }

    def retrieve_fields(self):
        if self.low is not None or self.high is not None:
            yield qualified_name(self.fieldname)

    def form_key(self):
        return self.fieldname


def choice_sql_expr(qualified_name, values, multiselect=False):
    """Construct SqlExpr to test whether a field has one of many values.
    This implements the main logic used by the Choice class when
    creating the SqlExpr. It is a separate function to make it easier to
    reuse it in special Expr sub-classes that are similar to Choice,
    but different enough not to be derived from Field.

    If multiselect is true (default: false), the field is assumed to
    contain an array and the generated SQL tests for overlap of the
    field and the values array.
    """
    if values is None:
        return E.TRUE
    if values:
        if not multiselect:
            return E.EQ(E.Field(qualified_name), E.ArrayAny(E.Const(values)))
        else:
            return E.Overlaps(E.Field(qualified_name), E.Const(values))
    return E.FALSE


class Choice(Expr):

    """Search condition for choice fields defined by the formed tree.
    The condition implemented by this class matches an integer field
    against a list of values.
    """

    def __init__(self, fieldname, values, multiselect=False):
        """Initialize the Choice condition.

        The parameter fieldname should be a string with the unqualified
        fieldname (for the generated SQL, a "m." prefix will be added).

        The parameters values describes the values to match agains.
        There are three cases to consider:

          1. values is None.

             This means the value of the field doesn't matter and thus
             the generated SQL will simply be 'TRUE'.

          2. values is an empty list.

             This means no value matches and no case will match. The
             generated SQL will simply be 'FALSE'

          3. The value is a list.

             In this case the generated SQL is a boolean expression that
             evaluates to TRUE if the value of the field is one of the
             values in the list.

        The parameter multiselect indicates whether the field in the
        database can have multiple values and is represented by an
        array. For most fields this is not the case, so the default for
        multiselect is False.
        """
        self.fieldname = fieldname
        self.values = values
        self.multiselect = multiselect

    def __repr__(self):
        return ("Choice(%r, %r, multiselect=%r)"
                % (self.fieldname, self.values, self.multiselect))

    def sql_expr(self):
        return choice_sql_expr(qualified_name(self.fieldname), self.values,
                               multiselect=self.multiselect)

    def to_simple_format(self):
        return {
            'cls': self.__class__.__name__,
            'attrs': {"fieldname": self.fieldname,
                      "values": self.values,
                      "multiselect": self.multiselect,
                      },
        }

    def retrieve_fields(self):
        if self.values is not None and self.values:
            yield qualified_name(self.fieldname)

    def form_key(self):
        return self.fieldname


class Radio(Field):

    def __repr__(self):
        return "Radio(%r, %r)" % (self.fieldname, self.value)

    def sql_expr(self):
        if self.value == -1:
            return E.NAryOr([E.EQ(E.Field(qualified_name(self.fieldname)),
                E.Const(-1)),
                E.IS_NULL(E.Field(qualified_name(self.fieldname)))])
        return field_sql_expr(qualified_name(self.fieldname), E.EQ,
                              self.value)

class Responsibilities(Expr):

    """Search condition that matches based on owner and standin.
    """

    def __init__(self, own, standin, changed_editor=False):
        """Initialize the search condition.
        The parameters own, standin and changed_editor are treated as booleans.

        The parameter changed_editor has a default value for backwards
        compatibility reasons: there might be saved searches that do not
        contain this parameter.

        If the user executing the query is an administrative user, the
        SQL generated for this condition is simply "TRUE".

        Otherwise the semantics depends on the values of own, standin
        and changed_editor. Each by itself has the following meaning:

           "own": Match cases for which the current user is the owner
                  (user-ID == master_tbl.bearbeiter_id)

           "standin": Match cases which the user can see/edit because
                      they're standins for the actual owner. Since a
                      user cannot be standin for a case they own, this
                      is equivalent to not being the owner.

           "changed_editor": Match cases which at some point in the past
                             have been transferred from user to another
                             (st.changed_editor is true)

        Obviously, "own" and "standin" are mutually exclusive, i.e.
        there cannot be a case where a user is both owner and standin
        for a case.

        This class interprets the user's intention in the following
        ways:

        We can consider "own" and "standin" together but separately from
        "changed_editor".

        When exactly one of "own" and "standin" is true, the user's
        intentation here is assumed to be as described above. When
        neither or both are true, this class generates SQL that does not
        check for owner or standin at all. The reasoning is that if
        neither are selected, the user expresses that this criterion
        should not be considered, and if both are selected it means
        showing both the user's own cases and the standin cases
        (effectively OR-ing the conditions). Using AND instead is not
        useful because the query result would be always be empty.

        "changed_editor", if true, is interpreted as described above. If
        it's false, the "changed editor" criterion is not used in the
        query. Like above, not checking the checkbox is interpreted as
        "disregard this criterion in the search".

        This leaves the question how the criteria for "own"/"standin" on
        the one hand and "changed_editor" on the other are combined, if
        both are to be applied in the search. This class chooses to
        again use OR instead of AND for consistency.
        """
        self.own = own
        self.standin = standin
        self.changed_editor = changed_editor

    def __repr__(self):
        return ("Responsibilities(%r, %r, %r)"
                % (self.own, self.standin, self.changed_editor))

    def sql_expr(self):
        if hasRole(['admin', 'pb']):
            return E.TRUE

        user = session['USER_AUTHORIZED']

        conditions = []
        if self.own and not self.standin:
            conditions.append(E.EQ(E.Field("m.bearbeiter_id"),
                                   E.Const(user.id)))
        elif not self.own and self.standin:
            conditions.append(E.NE(E.Field("m.bearbeiter_id"),
                                   E.Const(user.id)))

        if self.changed_editor:
            conditions.append(E.Field("st.changed_editor"))

        if not conditions:
            conditions.append(E.TRUE)

        return E.NAryOr(conditions)

    def to_simple_format(self):
        return {
            'cls': self.__class__.__name__,
            'attrs': {"own": self.own,
                      "standin": self.standin,
                      "changed_editor": self.changed_editor,
                      },
        }

    def form_key(self):
        return "responsibilities"


class PhaseDates(Expr):

    def __init__(self, phases, start_date, end_date):
        self.phases = phases
        self.start_date = start_date
        self.end_date = end_date

    def __repr__(self):
        return ("PhaseDates(%r, %r, %r)"
                % (self.phases, self.start_date, self.end_date))

    def to_simple_format(self):
        return {
            'cls': self.__class__.__name__,
            'attrs': {"phases": self.phases,
                      "start_date": self.start_date,
                      "end_date": self.end_date,
                      },
        }

    def sql_expr(self):
        sd = self.start_date
        ed = self.end_date

        if sd is None or ed is None:
            # no date range given -> match according to current phase
            return Choice("phase", self.phases).sql_expr()

        phases = self.phases
        phase_clauses = []
        sm = g.case_factory.get_state_machine()
        for p_id, des in sm.get_search_items():
            datefields = sm.get_state_datefield_names(p_id)
            ids = sm.get_state_ids_from_search_item(p_id)

            sp = datefields[0]
            ep = datefields[1]
            sp_id = ids[0]
            ep_id = ids[1]

            if sp is None or ep is None:
                # ignore phases without datefields.  This is usually
                # only the 'unknown' phase.
                continue

            # if phases is None, it means the phases parameter was
            # omitted altogether and all phases should match.
            if phases is None or sp_id in phases or ep_id in phases:
                # a phase can either be finished, in which case we have
                # both start and end dates for the phase, or it may be
                # still running, in which case we have a start date but
                # no end date yet.  This means that cases where the
                # start date of the phase has not been set yet should
                # not match.  However, cases where the end date of the
                # phase has not been set yet should match.
                successors = sm.get_successor_state_ids(p_id)
                value_succ = E.Const([int(suc) for suc in successors])

                phase_clauses.append(
                    E.NAryAnd([E.IS_NOT_NULL(E.Field(sp)),
                               E.LE(E.Field(sp), E.Const(ed)),
                               E.GE(E.Coalesce([E.Field(ep), E.NOW]),
                                    E.Const(sd)),
                               E.EQ(E.Field("phase"), E.ArrayAny(value_succ)),
                               ]))

        # If phases includes -1 then add check for cases with unknown
        # phase.
        if phases is None or -1 in phases:
            phase_clauses.append(E.EQ(E.Field("phase"), E.Const(-1)))

        if phase_clauses:
            return E.NAryOr(phase_clauses)
        return E.FALSE

    def form_key(self):
        return "phase"


class FederalState(Expr):

    def __init__(self, federal_states):
        self.federal_states = federal_states

    def __repr__(self):
        return "FederalState(%r)" % (self.federal_states,)

    def to_simple_format(self):
        return {
            'cls': self.__class__.__name__,
            'attrs': {"federal_states": self.federal_states,
                      },
        }

    def sql_expr(self):
        if self.federal_states in (None, ""):
            return E.TRUE

        return E.Templated("m.fkz::bigint in "
                           "(SELECT id FROM fkz_tbl"
                           " WHERE bundesland_id = ANY ({0}))",
                           E.Const(self.federal_states))


class SpecialField(Expr):

    """Base class for non-formed fields that are still treated similarly.
    """

    qualified_name = None

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return "SpecialField<%r>(%r)" % (self.qualified_name, self.value)

    def to_simple_format(self):
        return {
            'cls': self.__class__.__name__,
            'attrs': {"value": self.value},
        }

    def sql_expr(self):
        return field_sql_expr(self.qualified_name, E.EQ, self.value)

    def retrieve_fields(self):
        if self.value is not None and self.value != "":
            yield self.qualified_name


class Editor(Choice):

    fieldname = "bearbeiter_id"
    multiselect = False

    def __init__(self, values):
        self.values = values

    def __repr__(self):
        return "Editor(%r)" % (self.values)

    def to_simple_format(self):
        return {
            'cls': self.__class__.__name__,
            'attrs': {"values": self.values},
        }


class Branch(SpecialField):

    qualified_name = "b.filiale"


class Status(Expr):

    qualified_name = "st.status"

    def __init__(self, values):
        self.values = values

    def __repr__(self):
        return "Status(%r)" % (self.values,)

    def to_simple_format(self):
        return {
            'cls': self.__class__.__name__,
            'attrs': {"values": self.values},
        }

    def sql_expr(self):
        return choice_sql_expr(self.qualified_name, self.values)

    def form_key(self):
        return "status"


class SpecialBetween(Expr):

    qualified_name = None

    def __init__(self, low, high):
        self.low = low
        self.high = high

    def __repr__(self):
        return ("SpecialBetween<%r>(%r, %r)"
                % (self.qualified_name, self.low, self.high))

    def to_simple_format(self):
        return {
            'cls': self.__class__.__name__,
            'attrs': {"low": self.low,
                      "hight": self.high,
                      },
        }

    def sql_expr(self):
        return between_sql_expr(self.qualified_name, self.low, self.high)

    def retrieve_fields(self):
        yield self.qualified_name


class Updated(Expr):

    qualified_name = "st.zugriff"

    def __init__(self, value, op=E.EQ):
        self.op = op
        self.value = value

    def __repr__(self):
        return "Updated(%r, %r)" % (self.value, self.op)

    def to_simple_format(self):
        return {
            'cls': self.__class__.__name__,
            'attrs': {"op": self.op.__name__,
                      "value": self.value,
                      },
        }

    def sql_expr(self):
        # convert the value of the field to a date because the user can
        # only input dates and queries for all cases last modified on
        # that date shuld work.
        return field_sql_expr(self.qualified_name + "::date", self.op,
                              self.value)

    def retrieve_fields(self):
        if self.value is not None and self.value != "":
            yield self.qualified_name

    def form_key(self):
        return "updated"




class Needle(Expr):

    """Search condition for the main text search field.

    The test consists of a substring match against a number of fields
    given by the qualified_fields parameter which is a list of qualified
    field names (i.e. names including e.g. the table name).
    """

    def __init__(self, qualified_fields, pattern):
        self.qualified_fields = qualified_fields
        self.pattern = pattern

    def __repr__(self):
        return "Needle(%r, %r)" % (self.qualified_fields, self.pattern)

    def to_simple_format(self):
        return {
            'cls': self.__class__.__name__,
            'attrs': {"qualified_fields": self.qualified_fields,
                      "pattern": self.pattern,
                      },
        }

    def sql_expr(self):
        if not self.pattern or not self.qualified_fields:
            return E.TRUE
        pattern = E.Const(self.pattern)
        return E.NAryOr([E.SubstringInsensitive(E.Field(field), pattern)
                         for field in self.qualified_fields])


    def form_key(self):
        return "search_str"


class And(Expr):

    """Combine multiple conditions to one condition with AND."""

    def __init__(self, exprs):
        self.exprs = exprs

    def __repr__(self):
        return "And(%r)" % (self.exprs,)

    def retrieve_fields(self):
        for expr in self.exprs:
            for field in expr.retrieve_fields():
                yield field

    def sql_expr(self):
        if self.exprs:
            return E.NAryAnd([e.sql_expr() for e in self.exprs
                              if e is not TRUE])
        return E.TRUE

    def to_simple_format(self):
        return {'cls': 'And',
                'attrs': {
                    'exprs': [expr.to_simple_format() for expr in self.exprs]
                    }
                }

    def add_formed_field(self, formedtree, name):
        """Add the formedtree field named name to the and expression.
        This method does not modify the And-expression in place. It
        returns a new expression, unless there already is an expression
        for the given field, in which case the And expression is
        returned unchanged.
        """
        for expr in self.exprs:
            if expr.form_key() == name:
                return self
        return And(self.exprs + [formed_field_expr(formedtree, name)])

    def add_special_field(self, name):
        """Similar to formed fields but working with special fields.
        Special fields are fields that are not part of the formed
        tree like state or editor."""
        for expr in self.exprs:
            if expr.form_key() == name:
                return self
        return And(self.exprs + [special_field_expr(name)])

    def delete_form_field(self, key):
        """Return a new And expression with the named key removed."""
        return And([expr for expr in self.exprs if expr.form_key() != key])


def to_simple_format_from_json(json_tree):
    return json.loads(json_tree, object_hook=json_deserializer)


def to_searchtree(simple_format_tree):
    # first handle the special cases: And, because it is not a leaf
    #                                 TRUE, because use of singleton
    #                                 Field, because of the op
    cls = simple_format_tree['cls']
    attrs = simple_format_tree.get('attrs')
    if cls == 'And':
        all_exprs = attrs.get('exprs')
        filtered_exprs = [expr for expr in all_exprs
                          if expr["cls"] != "ChangedEditor"]
        if len(filtered_exprs) < len(all_exprs):
            # At least one item in all_exprs was a ChangedEditor expr.
            # Find the Responsibilities expr, if any, and set its
            # changed_editor attribute, otherwise, add a
            # Responsibilities expr
            for expr in filtered_exprs:
                if expr["cls"] == "Responsibilities":
                    expr["attrs"]["changed_editor"] = True
                    break
            else:
                filtered_exprs.append({"cls": "Responsibilities",
                                       "attrs": {"own": False, "standin": False,
                                                 "changed_editor": True}})
        return And([to_searchtree(expr) for expr in filtered_exprs])
    elif cls == 'TRUE':
        return TRUE
    else:
        if cls in ('Field', 'DayTimestamp', 'Updated'):
            attrs["op"] = getattr(E, attrs['op'])

        # Fix pre-multi-select searches for logbook art.
        if (cls == "Choice" and attrs["fieldname"] == "art"
            and "multiselect" not in attrs):
            attrs["multiselect"] = True

        try:
            return getattr(sys.modules[__name__], cls)(**attrs)
        except AttributeError:
            # Look for application specific Expressions
            app_module = import_overridable_module("lib.searchexpr")
            return getattr(app_module, cls)(**attrs)


def searchexpr_from_json(json):
    """Decode the given JSON representaion of a searchtree."""
    return to_searchtree(to_simple_format_from_json(json))


def formed_field_expr(formedtree, name):
    """Return a new Expr that matches a formedtree field.
    The defaults for e.g. comparison operators depend on the field type.
    """
    nc = formedtree.name2nc[name]

    if isinstance(nc, (formed.meta.data.DateLeaf,
                       formed.meta.data.IntLeaf,
                       formed.meta.data.NumericLeaf,
                       formed.meta.data.TextLeaf,
                       formed.meta.data.TextAreaLeaf)):
        return Field(name, None)
    elif isinstance(nc, formed.meta.data.ChoiceNode):
        return Choice(name, None)
    elif isinstance(nc, formed.meta.data.RadioNode):
        return Radio(name, None)
    elif isinstance(nc, formed.meta.data.BoolLeaf):
        return Bool(name, None)
    else:
        ValueError("Cannot create searchexpr for formed field %r" % (nc,))


def special_field_expr(name):
    """ Return an expression for a special field. See
    searchfield.py for a definition of the field names."""
    if name == "search_str":
        # TODO: With the more-roles branch this will be moved
        # into a single call to the config.
        user = session['USER_AUTHORIZED']
        if user.isAdmin():
            cfg = g.mpuls_config.get('search', 'admin_defaults')[0]
        else:
            cfg = g.mpuls_config.get('search', 'cm_defaults')[0]
        return Needle(cfg.get("needle"), None)
    if name == "status":
        return Status(None)
    if name == "responsibilities":
        return Responsibilities(False, False, False)
    if name == "phase":
        # FIXME: Using Choice here is most likely not correct. We could
        # have this without special treatment of "phase" by simply
        # handling it liek any other field in the formed tree. We most
        # likely want PhaseDates instead but that isn't implemented
        # properly in the user interface. So, using Choice a really just
        # a simple solution for a defect (previously the code here tried
        # to use a non-existant Phase class) that is at least good
        # enough at the time of the fix because it matches what a newly
        # logged in user sees as the default search criteria: a Choice
        # node for "phase".
        # See mpuls/issue6030, mpuls/issue6049
        return Choice("phase", None)
    if name == "bearbeiter_id":
        return Editor(None)
    if name == "federal_state":
        return FederalState(None)
    if name == "id":
        return Field("id", None)
    if name == "updated":
        return Updated(None)
    if g.mpuls_config.is_enabled("case-module", "logbook"):
        if name == "art":
            return Choice(name, None, multiselect=True)
        if name == "datum":
            return DayTimestamp(name, None)
        if name in LOGBOOK_EXTRAS:
            return Field(name, None)
    raise RuntimeError("Called with unknown name: %s" % name)
