#!/bin/sh

set -e

usage()
{
cat << EOF
usage: $0 options

This script will export databases in a sqlite database file from a given cluster.
Which databases are about to be exported is defined in.

    a) The Sqlite-template file and
    b) Port of the db-cluster.

The script will the export all database defined in the templatefile which can
be found on the given cluster and stores it as bz2 file under
"/home/mpuls/LOALNAME/transfer/data".

OPTIONS:
   -h      Show this message
   -d      Working directory. Default: current dir. 
   -c      Export commend to use. Default: export-sqlite.py
   -f      SQLite template file used for the export. Default: template.db
   -l      Local name of the export
   -p      Port of the databasecluster to export
   -t      Comma-separated list of tables to export. For details on the
           format, see the --relations option of export-sqlite.py
   -F      Comma-separated list of tables with fkz columns to be filled
           during export. See the --fkz-relations option of
           export-sqlite.py
   -b      List of fields that should not be exported. This is passed to
           export-sqlite.py with the --blacklist option.
   -j      Name of the application (e.g waska). Used to call the calc_jobperc
           script. Calc_jobperc is only called if this option is set.
   -n      Don't export logs and tags.
   -v      Verbose (unused)
EOF
}

# DEFAULTS
VERBOSE=0
EXPORTCOMMAND=export-sqlite.py
TYPE=
RTYPE=
DBPORT=
RELATIONS=""
FKZ_RELATIONS=""
TEMPLATE=template.db
BLACKLIST=""
BASEDIR=`pwd`
JOBPERC=""
NO_LOGS_TAGS=""

# Parse Options
while getopts "hb:vc:l:p:f:F:t:d:j:n" OPTION
do
        case $OPTION in
            h)
                usage
                exit 1
                ;;
	    b)
		BLACKLIST=$OPTARG
		;;
            c)
                EXPORTCOMMAND=$OPTARG
                ;;
            l)
                TYPE=$OPTARG
                ;;
            f)
                TEMPLATE=$OPTARG
                ;;
            p)
                DBPORT=$OPTARG
                ;;
            t)
                RELATIONS=$OPTARG
                ;;
            F)
                FKZ_RELATIONS=$OPTARG
                ;;
            d)
                BASEDIR=$OPTARG
                ;;
            v)
                VERBOSE=1
                ;;
            j)
                JOBPERC=$OPTARG
                ;;
            n)
                NO_LOGS_TAGS="--no_logs_tags"
                ;;
            ?)
                # Will be set if the optstring begins with a ":". In this case
                # all options not defined in the optstring will OPTION set to
                # "?". Otherwise the shell reports an error on the unknown
                # option.
                usage
                exit
                ;;
        esac
done

# Check Options
if [ -z $TYPE ]
then
    echo "Error! Name for export required!";
    usage;
    exit 1;
fi

if [ -z $DBPORT ]
then
    echo "Error! Port of the database to export is required!";
    usage;
    exit 1;
fi
if [ -z $TEMPLATE ]
then
    echo "Error! Template File for the export is required!";
    usage;
    exit 1;
fi

# Finally excecute program
USER=mpuls

DATADIR=${BASEDIR}/data

COMMAND=${BASEDIR}/${EXPORTCOMMAND}
echo ${COMMAND}
DATE=`date +"%Y-%m-%d"`
EXPORTFILE=${DATADIR}/${TYPE}-${DATE}.db

# Copy template file for a new export and export the data.
echo -- Start Export ${DATE} --
cp ${TEMPLATE} ${EXPORTFILE}
chmod 777 ${DATADIR}
chmod 666 ${EXPORTFILE}
sudo -u postgres ${COMMAND} --relations="${RELATIONS}" \
    --fkz-relations="${FKZ_RELATIONS}" \
    --blacklist="${BLACKLIST}" -p ${DBPORT} -o ${EXPORTFILE} ${NO_LOGS_TAGS}
if [ ${JOBPERC} ]
then
    sudo -u postgres ${BASEDIR}/helpers/jp/calc_jp.py -p ${DBPORT} -o ${EXPORTFILE} \
        -s "2011-09-01" -m "36" -w ${JOBPERC}
fi
bzip2 -f ${EXPORTFILE}
chmod 640 ${EXPORTFILE}.bz2
chmod 755 ${DATADIR}
echo -- End Export ${DATE} --

exit 1
