#! /usr/bin/python
# -*- coding: utf-8 -*-

"""This script takes one of the usual evaluation configuration files as
parameter and quite a few more options for the database connection and
some other evaluation parameters which are usually hard-wired somewhere
else (in the other evalation scripts in the scripts sub directory or in
the JSON-configuration of the webapplication (phases, date start/end
fields), performs the evaluation and writes the result in one of a
number of different formats (XML, CSV, XLS).
"""

import sys
from optparse import OptionParser
import getpass
import datetime

from libmpuls.evaluation.config import EvaluationConfig
from libmpuls.evaluation.evaluation import EvaluationSet
from libmpuls.evaluation.export import EvaluationExportXLS, \
     EvaluationExportCSV, EvaluationExportXML

from libmpuls.db.db import get_db


# known export formats.
export_format_map = {
    "xls": EvaluationExportXLS,
    "csv": EvaluationExportCSV,
    "xml": EvaluationExportXML,
    }


def load_evaluation(evaluation_filename, db, phases, year,
                    start_field, end_field):
    config = EvaluationConfig(evaluation_filename, db,
                              datetime.date(year, 1, 1).isoformat(),
                              datetime.date(year, 12, 31).isoformat(),
                              start_field, end_field, phases,
                              "SELECT %(fields)s"
                              "  FROM master_tbl_eval_total_view m"
                              "  JOIN ka_status_tbl_view st"
                              "    ON st.master_id = m.id")
    if not hasattr(config, "xml_doc"):
        # In the XML file cannot be loaded and parsed, EvaluationConfig
        # prints an error message to stdout, but does not indicate
        # properly to the caller that it couldn't read the
        # configuration. AFICT, the only indicator is that config does
        # not have the xml_doc attribute which would hold the etree
        # instance.
        raise IOError("Could not load evaluation configuration file %r"
                      % (evaluation_filename,))
    return config


def run_evaluation(config, exporter):
    evaluationset = EvaluationSet(config, False)
    evaluationset.evaluate()
    return evaluationset.export(exporter())


def option_parser():
    parser = OptionParser()
    parser.add_option("--host", default="localhost",
                      help="Host-name for the database connection")
    parser.add_option("--port", type="int", default=5432,
                      help="Port number for the database connection")
    parser.add_option("--db",
                      help="Database name for the database connection")
    parser.add_option("--user",
                      help="User name for the database connection")
    parser.add_option("--password", default=None,
                      help=("Password for the database connection. If omitted,"
                            " the program will prompt for a password"))
    parser.add_option("--evaluation",
                      help=("name of the XML file with the evaluation"
                            " specification"))
    parser.add_option("--phases", default="2,3",
                      help="comma separated list of phases (default: 2,3)")
    parser.add_option("--year", type="int",
                      help="The year to evaluate (default: current year)")
    parser.add_option("--start-field", default="datum_cm_start",
                      help=("Field (sql expression) for the field containing"
                            " the start date for the phases."))
    parser.add_option("--end-field",  default="datum_cm_ende",
                      help=("Field (sql expression) for the field containing"
                            " the end date for the phases."))
    parser.add_option("-o", "--output", help="name of the output file")
    parser.add_option("-f", "--format", type="choice",
                      choices=sorted(export_format_map.keys()),
                      default="xls", help="format of the output file")
    return parser


def parse_and_check_arguments():
    """Parse and check the command line parameters.

    The return value is a pair of (options, args) as returned by
    OptionParser.parse_args(). This function performs some plausibility
    checks on the options and arguments. Any errors found are printed to
    stderr. If any errors were found, sys.exit is called with exit code 1.
    """
    options, args = option_parser().parse_args()

    options_ok = True
    if options.db is None:
        print >> sys.stderr, "Error: A database must be given (--db)"
        options_ok = False
    if options.user is None:
        print >> sys.stderr, ("Error: A user name must be given"
                              " (--user)")
        options_ok = False
    if options.evaluation is None:
        print >> sys.stderr, ("Error: An evaluation XML file must be given"
                              " (--evaluation)")
        options_ok = False
    if options.output is None:
        print >> sys.stderr, ("Error: An output file must be given"
                              " (--output, -o)")
        options_ok = False
    if not options_ok:
        sys.exit(1)

    return (options, args)


def main():
    options, args = parse_and_check_arguments()

    exporter = export_format_map[options.format]

    if options.year is None:
        options.year = datetime.date.today().year

    if options.password is None:
        options.password = getpass.getpass("Password for %r: " % options.user)

    db = get_db(options.user, options.password, options.db, options.host,
                options.port)
    con = db.getConnection()
    try:
        config = load_evaluation(options.evaluation, con, options.phases,
                                 options.year, options.start_field,
                                 options.end_field)
        result = run_evaluation(config, exporter)
    finally:
        db.recycleConnection(con)
        db.closeConnections()

    if isinstance(result, unicode):
        result = result.encode("utf-8")

    with open(options.output, "w") as out:
        out.write(result)



main()
