# -*- coding: utf-8 -*-
# Copyright 2009 Intevation GmbH, Germany, <info@intevation.de>
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
#

IDENTIFIER_SEPARATOR = u':'


class InstanceComponent(object):

    def __init__(self, meta=None):
        self.meta = meta
        self.parent = None
        self.tree = None
        self.identifier = None
        self.external_identifier = None

    def isRepeat(self):
        return False

    def isLeaf(self):
        return False

    def getParent(self):
        return self.parent

    def getMeta(self):
        return self.meta

    def getInstanceTree(self):
        if self.tree:
            return self.tree
        cur = self.parent
        tree = None
        while cur:
            tree = cur.getInstanceTree()
            if tree:
                break
            cur = cur.parent
        self.tree = tree
        return tree

    def getIdentifier(self):
        if self.external_identifier is not None:
            return self.external_identifier

        path = [self.getMeta().getName()]
        cur = self
        while cur:
            if cur.isRepeat():
                path.append(str(cur.getInternalIdentifier()))
            cur = cur.parent
        identifier = IDENTIFIER_SEPARATOR.join(path)
        self.external_identifier = identifier
        return identifier

    def getInternalIdentifier(self):
        return None

    def hasErrors(self):
        '''Returns True if the fields has an error'''
        tree = self.getInstanceTree()
        if tree:
            return tree.hasErrors(self.getIdentifier())
        return False

    def getErrors(self):
        '''Returns Error message of the field as str'''
        tree = self.getInstanceTree()
        if tree:
            return tree.getErrors(self.getIdentifier())
        return None

    def hasWarning(self):
        '''Returns True if the fields has a warning'''
        tree = self.getInstanceTree()
        if tree:
            return tree.hasWarning(self.getIdentifier())
        return False

    def getWarning(self):
        '''Returns Warning message of the field as str'''
        tree = self.getInstanceTree()
        if tree:
            return tree.getWarning(self.getIdentifier())
        return None


class InstanceLeaf(InstanceComponent):

    def __init__(self, meta=None, value=None):
        super(InstanceLeaf, self).__init__(meta)
        self.value = value

    def getValue(self):
        return self.value

    def setValue(self, value):
        self.value = value

    def isLeaf(self):
        return True


class InstanceNode(InstanceComponent):

    def __init__(self, meta=None, children=None, nodes=None):
        super(InstanceNode, self).__init__(meta)
        self.children = children
        self.nodes = nodes
        if self.nodes is None:
            self.nodes = {}

    def getChildren(self):
        return self.children

    def hasErrors(self):
        my_ids = (IDENTIFIER_SEPARATOR +
                  self.getIdentifier().split(IDENTIFIER_SEPARATOR, 1)[1])
        tree = self.getInstanceTree()
        find = tree.getFormEd().findByName
        my_meta = self.getMeta()
        for identifier in tree.getErrors().iterkeys():
            iname, ids = identifier.split(IDENTIFIER_SEPARATOR, 1)
            if not (IDENTIFIER_SEPARATOR + ids).endswith(my_ids):
                continue
            meta = find(iname)
            while meta:
                if meta == my_meta:
                    return True
                meta = meta.parent
        return False

    def getErrors(self):
        my_name, my_idnr = self.getIdentifier().split(IDENTIFIER_SEPARATOR, 1)
        my_ids = IDENTIFIER_SEPARATOR + my_idnr
        errors = []
        tree = self.getInstanceTree()
        find = tree.getFormEd().findByName
        my_meta = self.getMeta()
        for identifier, error_item in tree.getErrors().iteritems():
            iname, ids = identifier.split(IDENTIFIER_SEPARATOR, 1)
            if not (IDENTIFIER_SEPARATOR + ids).endswith(my_ids):
                continue
            meta = find(iname)
            while meta:
                if meta == my_meta:
                    errors.append(error_item)
                    break
                meta = meta.parent

            # Now handle references of the iname
            refs = tree.getFormEd().name2ref.get(iname, [])
            for ref in refs:
                meta = ref
                while(meta):
                    if meta.getName() == my_name:
                        errors.append(error_item)
                        break
                    meta = meta.parent
        return errors

    def get_node(self, name):
        return self.nodes[name]

    def set_node(self, name, value):
        self.nodes[name] = value

    def has_node(self, name):
        return name in self.nodes


class PageInstanceNode(InstanceNode):

    pass


class RepeatInstanceContainerNode(InstanceNode):

    def createInstance(self, uuid=None):
        return None

    def deleteInstance(self, instance_node):
        pass


class RepeatInstanceNode(InstanceNode):

    def __init__(self, meta=None, children=None, identifier=None):
        super(RepeatInstanceNode, self).__init__(meta, children)
        self.identifier = identifier

    def isRepeat(self):
        return True


class RootNode(InstanceNode):

    def __init__(self, meta=None, children=None, identifier=None):
        super(RootNode, self).__init__(meta, children)
        self.identifier = identifier

    def getIdentifier(self):
        return self.identifier

    def isRepeat(self):
        return True


class SetDataException(Exception):

    def __init__(self, errors):
        self.errors = errors
        Exception.__init__(self, self.format_message())

    def format_message(self):
        """Return a formatted representation for logging purposes.
        The values that cause the errors are not contained in the result
        to avoid leaking sensitive data.
        """
        return repr([(name, item.type, item.msg)
                     for name, items in self.errors.items() for item in items])


class InstanceTree(object):

    def __init__(self, formed, root=None):
        self.formed = formed
        if root is None:
            root = RootNode()
        self.root = root

    def getFormEd(self):
        return self.formed

    def getItem(self, identifier, name=None):
        return None

    def getAllItemsForName(self, name):
        return None

    def hasErrors(self, identifier=None):
        return None

    def getErrors(self, identifier=None):
        return None

    def commit(self):
        return None

    def setData(self, data):
        pass

    def setDataNoChecks(self, data):
        pass

    def getRootNode(self):
        return self.root

    def get_node(self, name):
        return self.root.get_node(name)

    def has_node(self, name):
        return self.root.has_node(name)


# vim:set ts=4 sw=4 si et sta sts=4 :
