# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 by Intevation GmbH
#
# FormEd is free software under the terms of the
# GNU General Public License Version 3. See file
# LICENSE coming with the source of FormEd for details.
#
# Authors:
# Bernhard Herzog <bernhard.herzog@intevation.de>

"""Babel Message Extractor for FormEd Meta Trees.

With the entry point defined in libformed's setup.py file, you can use
sections like the following to define the extractors in the babel config
file:

   [formedtree: formedtree.xml]
"""

from formed.meta.io.document import openDocument
from formed.meta.nodecomponents import Node


def extract_formedtree(fileobj, keywords, comment_tags, options):
    """Extract messages from formedtree XML files.
    This function only uses its first argument, a file-like object from
    which the formed tree should be read. The other arguments exist for
    compatibility with babel.

    The file-like object must have an attribute "name" containing the a
    filename suitable for re-opening the file. The reason for this is
    that libformed needs to be able to locate other files referenced by
    the formedtree.xml file like external choice lists and annotations
    and it does so based on the file-name. However, babel only provides
    a file object and not the file-name. In order to work around this,
    we rely on an implementation detail of babel: It passes a normal
    python file object created by calling open with the file-name we
    want to know. As a consequence that filename is available as the
    name attribute on the file-object.

    The return value is an iterable containing (lineno, funcname,
    message, comments) tuples.
    """
    meta = openDocument(fileobj.name)
    for msg in extract_from_nc(meta.root):
        yield msg


def extract_from_nc(nc):
    """Extract translatable strings from a node-component and its descendants.
    The translatable strings are all strings which might be presented to
    the user: description, title, placeholder, and all annotations.

    This function is a generator yielding tuples of the form required by
    babel. We do not have line-number information, so we always use 0 as
    the line-number. Also, there's no function marking the strings, so
    we always return an empty string as the function name. The comments
    element is a list of strings containing the node component name and
    the attribute or annotation name whose string forms the message
    element.
    """
    for msg in extract_attributes(nc, "description", "title", "placeholder"):
        yield msg

    annotations = nc.annotations
    if annotations:
        for ann in annotations:
            yield (0, "", ann.msg, ["%s.%s" % (nc.getName(), ann.kind)])

    if isinstance(nc, Node):
        for child in nc.getChildren():
            for msg in extract_from_nc(child):
                yield msg


def extract_attributes(nc, *attrs):
    for attr in attrs:
        value = nc.getAttribute(attr)
        if value:
            yield (0, "", value, ["%s.%s" % (nc.getName(), attr)])
