# -*- coding: latin-1 -*-
#
# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
#

VISIT_IGNORE_CHILDREN = 2
VISIT_CONTINUE = None

REQUIRED = ("required",)
HELP = ("help",)
EVALUATION = ("evaluation",)


class UnknownFlagError(Exception):

    def __init__(self, nc, flag, unparsed_value):
        self.nc = nc
        self.flag = flag
        self.unparsed_value = unparsed_value

    def __str__(self):
        return ("Unknown flag %r in NodeComponent %r" % (self.flag, self.nc))


class NodeComponent(object):

    default_flags = {
        'required': [],
        'time': [],
        'required-mark': 'warning',
        'evaluation': None,
    }
    flag_splitters = {
        "required": lambda x: x.split(","),
        "time": lambda x: x.split(","),
        "required-mark": lambda x: x,
        "evaluation": lambda x: x,
    }


    def __init__(self):
        self.parent = None
        self.attributes = {}
        self.attributes["name"] = ""
        self.attributes["description"] = ""
        self.attributes["title"] = ""
        self.attributes["formularname"] = ""
        self.attributes["modes"] = ""
        self.attributes["annotation"] = ""
        self.attributes["alternative"] = ""
        self.attributes["flags"] = ""
        self.attributes["function"] = ""
        self.attributes["span"] = ""

        self.found_default = None
        self.findDefault = self.firstFindDefault
        self.annotations = None
        self.parsed_flags = None

    def setAnnotations(self, annotations):
        self.annotations = annotations

    def hasAnnotations(self, kinds):
        '''Returns True if the field as an annotion (help, required...)'''
        if not self.annotations:
            return False
        for annotation in self.annotations:
            if annotation.kind in kinds:
                return True
        return False

    def getAnnotations(self, kinds):
        if not self.annotations:
            return u''
        return u''.join([a.msg for a in self.annotations if a.kind in kinds])

    def hasHelp(self):
        return self.hasAnnotations(HELP)

    def getHelp(self):
        return self.getAnnotations(HELP)

    def hasRequired(self):
        return self.hasAnnotations(REQUIRED)

    def getRequired(self):
        return self.getAnnotations(REQUIRED)

    def hasEvaluation(self):
        return self.hasAnnotations(EVALUATION)

    def getEvaluation(self):
        return self.getAnnotations(EVALUATION)

    def firstFindDefault(self):
        self.findDefault = self.defaultFound
        return self.found_default

    def defaultFound(self):
        return self.found_default

    def setAttribute(self, key, value):
        self.attributes[key] = value

    def getAttribute(self, key):
        return self.attributes.get(key, None)

    def getAnnotation(self):
        return self.getAttribute("annotation")

    def setAnnotation(self, annotation):
        return self.setAttribute("annotation", annotation)

    def getAlternative(self):
        return self.getAttribute("alternative")

    def setAlternative(self, alternative):
        return self.setAttribute("alternative", alternative)

    def setFormularName(self, name):
        self.setAttribute("formularname", name)

    def getFormularName(self):
        return self.getAttribute("formularname")

    def setName(self, name):
        self.setAttribute("name", name)

    def getName(self):
        return self.getAttribute("name")

    def setDescription(self, description):
        self.setAttribute("description", description)

    def getDescription(self):
        return self.getAttribute("description")

    def set_title(self, title):
        self.setAttribute("title", title)

    def getShortDescription(self):
        if self.getAttribute("short-description"):
            return self.getAttribute("short-description")
        else:
            return self.getAttribute("description")

    def get_title(self):
        if self.getAttribute("title"):
            return self.getAttribute("title")
        else:
            return self.getAttribute("description")

    def getTitle(self):
        return self.get_title()

    def setTitle(self, title):
        self.set_title(title)

    def getModes(self):
        return self.getAttribute("modes")

    def setModes(self, modes):
        self.setAttribute("modes", modes)

    def getFlags(self):
        return self.getAttribute("flags")

    def get_parsed_flags(self):
        """Return the flags attribute as a dictionary"""
        if self.parsed_flags is not None:
            return self.parsed_flags

        parsed_flags = self.default_flags.copy()
        flags = self.getFlags()
        if not flags:
            return parsed_flags

        for raw_flag in flags.split(';'):
            name, sep, values = raw_flag.partition(':')
            splitter = self.flag_splitters.get(name)
            if splitter is None:
                raise UnknownFlagError(self, name, values)
            parsed_flags[name] = splitter(values)

        self.parsed_flags = parsed_flags
        return parsed_flags

    def is_required(self):
        return bool(self.get_parsed_flags()["required"])

    def setFlags(self, flags):
        self.parsed_flags = None
        return self.setAttribute("flags", flags)

    def get_span(self):
        raw = self.getAttribute("span")
        if raw:
            return int(raw)
        return None

    def getParent(self):
        return self.parent

    def getFunction(self):
        return self.getAttribute("function")

    def pathToRoot(self):
        current = self
        path = []
        while current is not None:
            path.insert(0, current)
            current = current.parent
        return path

    def depth(self):
        d = -1
        current = self
        while current is not None:
            d += 1
            current = current.parent
        return d

    def walk(self):
        yield self

    def visit(self, visitor):
        return visitor(self)

    def shallowCopy(self):
        nc = self.__class__()
        nc.attributes = self.attributes.copy()
        return nc

    def clone(self):
        return self.shallowCopy()


class Node(NodeComponent):

    def __init__(self):
        super(Node, self).__init__()
        self.children = []

    def addChild(self, child, idx=-1):
        if idx == -1:
            idx = len(self.children)
        self.children.insert(idx, child)
        child.parent = self

    def indexOfChild(self, child):
        try:
            return self.children.index(child)
        except ValueError:
            return -1

    def getChildByName(self, name):
        for child in self.children:
            if child.getName() == name:
                return child
        return None

    def getChildren(self):
        return self.children

    def removeChild(self, child):
        try:
            self.children.remove(child)
            child.parent = None
        except ValueError:
            return

    def walk(self):
        yield self
        for child in self.children:
            for x in child.walk():
                yield x

    def visit(self, visitor):

        result = visitor(self)
        if result == VISIT_IGNORE_CHILDREN:
            return VISIT_CONTINUE
        if result != VISIT_CONTINUE:
            return result
        for child in self.children:
            result = child.visit(visitor)
            if result == VISIT_IGNORE_CHILDREN:
                return VISIT_CONTINUE
            if result != VISIT_CONTINUE:
                return result
        return VISIT_CONTINUE

    def clone(self):
        n = NodeComponent.clone(self)
        children = []
        for child in self.children:
            child = child.clone()
            children.append(child)
            child.parent = n
        n.children = children
        return n

    def firstLevel(self, crit):
        visited = set()
        for c in self.walk():
            if c != self and crit(c):
                parent = c.parent
                found = False
                while parent:
                    if parent in visited:
                        found = True
                        break
                    parent = parent.parent
                if not found:
                    visited.add(c)
                    yield c


class RootNode(Node):

    def __init__(self, document=None):
        super(RootNode, self).__init__()
        self.document = document
        self.attributes["version"] = ""
        self.attributes["boolean_choice_table"] = ""

    def getVersion(self):
        return self.getAttribute("version")

    def getBooleanChoiceTable(self):
        return self.getAttribute("boolean_choice_table")

    def getCaseNode(self):
        if self.children:
            return self.children[0]
        return None


class CaseNode(Node):

    pass


class Leaf(NodeComponent):

    def __init__(self):
        super(Leaf, self).__init__()
        self.attributes["value"] = ""

    def getValue(self):
        return self.getAttribute("value")

    def setValue(self, value):
        self.setAttribute("value", value)


# vim:set ts=4 sw=4 si et sta sts=4:
