# -*- coding: utf-8 -*-
#
# (c) 2008 by Intevation GmbH
# This is Free software under the GPLv3. 
# See LICENSE comming with the source of 'mpuls'
# for details.
#
# author: Torsten IrlÃ€nder <torsten.irlaender@intevation.de>

import logging

try:
    import xml.etree.ElementTree as etree
except ImportError:
    import elementtree.ElementTree as etree

EVAL_TBL  = """master_tbl_eval_total_view"""
TMP_SUB   = """tmp_subquery"""
TMP_SET   = """tmp_set"""
TMP_EVAL  = """tmp_eval"""
TMP_TABLE = """tmp_evaltbl"""
PREP_STMT = """
SELECT c.xid as xid, c.xvalue as x, c.yid as yid, c.yvalue as y, coalesce(%(function)s, %(function_coalesce_default)s) as result, count(DISTINCT m.id) as count
INTO TEMP TABLE %(tmp_tbl)s 
FROM 
( %(basequery)s ) m
RIGHT JOIN 
( SELECT x.value as xvalue, 
         x.id as xid, 
         y.value as yvalue,
         y.id as yid
         FROM
         %(xtable)s x, 
         %(ytable)s y
) c
ON 
m.%(xjoinon)s = c.xid
AND
m.%(yjoinon)s = c.yid
GROUP by c.xid, c.xvalue, c.yid, c.yvalue
"""
# Save cases matching the subquery into a temporary table
SET_SUB_STMNT = """SELECT DISTINCT m.* INTO TEMP TABLE %(tmp_tbl)s FROM ( %(subquery)s ) m"""

SET_FILTER_STMNT = """SELECT DISTINCT m.* INTO TEMP TABLE %(tmp_tbl)s FROM %(subquery)s m WHERE %(tags)s and %(where)s"""
EVAL_FILTER_STMNT = """SELECT DISTINCT m.* INTO TEMP TABLE %%(tmp_tbl)s FROM %s m WHERE %%(tags)s and %%(where)s""" % TMP_SET
FILTER_VALID_STMNT = """SELECT count(DISTINCT m.id)
FROM ( %(table)s ) m JOIN nm_tags_cases_view t ON m.id = t.case_id
WHERE %(tags)s AND %(where)s"""

log = logging.getLogger(__name__)

def build_map(path):
    '''Builds a mapping dictionary from file.
       File must have following format:

       old;new

       description will not be used here.
    '''

    f = None
    mapping = {}
    if not path: return
    try:
        try:
            f = open(path, 'r')
            for l in f:
                tokens = l.split(';')
                mapping[tokens[0].rstrip('\n')] = tokens[1].rstrip('\n')
            return mapping
        except StandardError, err:
            log.error(traceback.print_exc())
            log.error('Error on building mapfile: %s' % err)
            return {}
    finally:
        if f:
            f.close()

class EvaluationConfig:

    '''
    Configuration interface for evaluations
    '''

    def __init__(self, filename, conn, sdate, edate, sdatefield, edatefield,
                pphase, sql, elist=None, agency=None, mapping=None, sqlwhere=None,
                adele=False, sqlparams=None):
        '''
        @param filename: Path of the XML-config which defines the evaluations 
        @type filename: String
        @param sdate: Startdate of the evaluation period
        @type sdate: String as YYYY-MM-NN
        @param edate: Enddate of the evaluation period
        @type edate: String as YYYY-MM-NN
        @param sdatefield: Name of the dbfield which will be used for the startdate
        @type sdatefield: String
        @param edatefield: Name of the dbfield which will be used for the enddate
        @type edatefield: String
        @param pphase: Minimum phase which cases must have to be considered for evaluations
        @type pphase: Integer 
        @param sql: Complete subquery wich builds the basic set of cases which are used for evaluation.
        @type sql: String 
        @param sqlwhere: Userdefined sql where statement In contrast to the sql this can be used to add userdefined        search criteria.
        @type sqlwhere: String 
        @param agency: Name of the agency for which the evluation is done.
        @type agency: String. 
        @param elist: List of evaluation ids. The listed evaluations will be part of the evaluation.
        @type elist: Comma seperated list of integers. 
        @param mapping: Name of a CVS file with agencyname mappings.
        @type elist: String. 
        @param sqlparams: Additional parameter data which belong to sql
        @type sqlparams: Dictionary
        '''


        self.elist = [] 
        if elist is not None:
            self.elist = elist
        else:
            self.elist = elist
        
        


        self.conn       = conn

        self.agency     = agency 
        #Preconditions (user defined)
        self.sdate      = sdate
        self.edate      = edate
        self.sdatefield = sdatefield
        self.edatefield = edatefield
        self.pphase     = pphase

        self.sqlwhere   = sqlwhere 
        self.sql        = sql
        self.sqlparams  = sqlparams
        self.evaluation_filter = {}
        
        self.adele = adele

        self.mapping = None
        if mapping:
            try:
                self.mapping = build_map(mapping)
            except StandardError, err:
                print 'Could not mapping file: %s' % err

        try:
            xml          = open(filename, 'r')
            self.xml_doc = etree.parse(xml)
            xml.close()
        except StandardError, err:
            print 'Could not load xml evaluation config: %s' % err

    def set_filter(self, evaluation_filter):
        """
        Pass human-readable representations of the filters. This information
        will later be shown in the evaluations
        """
        self.evaluation_filter = evaluation_filter
    
    def _get_evaluation(self, id):
        '''Returns the configuration of a single evalutions'''
        for eval in self.xml_doc.findall('evaluation'):
            if id == int(eval.findtext('id')):
                return eval
        return None

    def get_required_setphase(self):
        '''
        Returns the required phase of the evaluationsset

        @return: id of the evaluationset 
        @rtype: String 
        '''
        return self.xml_doc.findtext('requirement/phase')

    def get_required_setvalidity(self):
        '''
        Returns the required validity of the evaluationsset

        @return: id of the evaluationset 
        @rtype: String 
        '''
        return [e.text for e in self.xml_doc.findall('requirement/validity')]

    def get_agency(self):
        '''
        Returns the name of the agency for which the evaluation is done. 
        If a mapping is configured try to map the given name to the new name.

        @return: Name of the agency. None if no name is defined
        @rtype: String 
        '''
        if self.mapping:
            return self.mapping.get(self.agency, self.agency)
        else:
            log.info('Could not find mapping for %s' % self.agency)
            return self.agency

    def get_version(self):
        '''
        Returns the version of the evaluationsset

        @return: version of the evaluationset 
        @rtype: String 
        '''
        return self.xml_doc.findtext('version')

    def get_setmode(self):
        '''
        Returns the mode of the evaluationsset id defined. Else None

        @return: mode of the evaluationset 
        @rtype: String or None
        '''
        try:
            return self.xml_doc.find('.').attrib['mode']
        except KeyError:
            return None
        return mode

    def get_setid(self):
        '''
        Returns the id of the evaluationsset

        @return: id of the evaluationset 
        @rtype: String 
        '''
        return self.xml_doc.findtext('id')

    def get_setname(self):
        '''
        Returns the name of the evaluationsset

        @return: name of the evaluationset 
        @rtype: String 
        '''
        return self.xml_doc.findtext('name')

    def get_setdesc(self):
        '''
        Returns the description of the evaluationsset

        @return: description of the evaluationset 
        @rtype: String 
        '''
        return self.xml_doc.findtext('desc')

    def get_evaluations(self):
        '''
        Returns a list of ids of available evaluations.

        @return: list of ids of available evaluations 
        @rtype: List of integers 
        '''
        ids = []
        for eval in self.xml_doc.findall('evaluation'):
            id = int(eval.findtext('id'))
            if (self.elist is None) or (str(id) in self.elist):
                ids.append(id)
        return ids

    def get_basequery(self, id): 
        '''
        Returns the basequery of the evaluation

        @param id: Id of the evaluation 
        @type id: Integer 
        @return: basequery of the evaluation
        @rtype: String
        '''
        config = self._get_evaluation(id)
        if config is None: return None
        return config.findtext('basequery/template')

    def get_basetable(self, id): 
        '''
        Returns the table of the evaluation

        @param id: Id of the evaluation 
        @type id: Integer 
        @return: table of the evaluation
        @rtype: String
        '''
        config = self._get_evaluation(id)
        if config is None: return None
        return config.findtext('basequery/table')

    def get_basewhere(self, id): 
        '''
        Returns the where part of the evaluation

        @param id: Id of the evaluation 
        @type id: Integer 
        @return: table of the evaluation
        @rtype: String
        '''
        config = self._get_evaluation(id)
        if config is None: return None
        return config.findtext('basequery/where')

    def get_baseaggregate(self, id): 
        '''
        Returns the where part of the evaluation

        @param id: Id of the evaluation 
        @type id: Integer 
        @return: aggregate fnction the evaluation default is count
        @rtype: String
        '''
        config = self._get_evaluation(id)
        if config is None: return None
        result = config.findtext('basequery/aggregate')
        return result or 'count(m.id)'

    def get_baseaggregatetotal(self, id):
        '''
        Returns the aggregate total of the basequery of evaluation id

        @param id: Id of the evaluation
        @type id: Integer
        @return: aggregatetotal of the basequery
        @rtype: String
        '''
        config = self._get_evaluation(id)
        if config is None: return None
        result = config.findtext('basequery/aggregatetotal')
        return result



    def get_name(self, id): 
        '''
        Returns the name of the evaluation

        @param id: Id of the evaluation 
        @type id: Integer 
        @return: Name of the evaluation
        @rtype: String
        '''
        config = self._get_evaluation(id)
        if config is None: return None
        return config.findtext('name')

    def get_short_name(self, id):
        '''
        Returns the short name of the evaluation.
        If no short name has been given, return the name that would be
        returned by get_name()

        @param id: Id of the evaluation
        @type id: Integer
        @return: Short name of the evaluation
        @rtype: String
        '''
        config = self._get_evaluation(id)
        if config is None: return None
        name = config.findtext('short_name')
        if name is None:
            name = self.get_name(id)
        return name

    def get_desc(self, id): 
        '''
        Returns the description of the evaluation

        @param id: Id of the evaluation 
        @type id: Integer
        @return: Description of the evaluation
        @rtype: String
        '''
        config = self._get_evaluation(id)
        if config is None: return None
        return config.findtext('desc')

    def get_xdims(self, id): 
        '''
        Returns the Label on X-Axis of the evaluation

        @param id: Id of the evaluation 
        @type id: Integer
        @return: X-Label of the evaluation
        @rtype: String
        '''
        config = self._get_evaluation(id)
        return config.findall('xdim')

    def get_ydims(self, id): 
        '''
        Returns the Label of the Y-Axis of the evaluation

        @param id: Id of the evaluation 
        @type id: Integer
        @return: Y-Label of the evaluation
        @rtype: String
        '''
        config = self._get_evaluation(id)
        return config.findall('ydim')

    def get_required_evalphase(self, id):
        '''
        Returns the required phase of the evaluationsset

        @param id: Id of the evaluation 
        @type id: Integer
        @return: id of the evaluationset 
        @rtype: String 
        '''
        config = self._get_evaluation(id)
        if config is None: return None
        return config.findtext('requirement/phase')

    def get_required_evalvalidity(self, id):
        '''
        Returns the required validity of the evaluationsset

        @param id: Id of the evaluation 
        @type id: Integer
        @return: id of the evaluationset 
        @rtype: String 
        '''
        config = self._get_evaluation(id)
        if config is None: return [] 
        return [e.text for e in config.findall('requirement/validity')]

    def get_evaluation_mode(self, id):
        '''
        Returns the mode of the evaluation
        '''
        config = self._get_evaluation(id)
        if config is None: return None
        return config.attrib.get('mode',None)

    def get_label(self, config):
        if config is None: return None
        return config.findtext('label')

    def get_table(self, config):
        if config is None: return None
        return config.findtext('table')

    def get_field(self, config):
        if config is None: return None
        return config.findtext('field')

    def get_join(self, config):
        if config is None: return None
        return config.findtext('join')

    def get_where(self, config):
        if config is None: return None
        return config.findtext('where') or "TRUE"

    def get_display(self):
        '''
        Return the display mode of the evaluations

        @return: Display mode of the evaluation
        @rtype: String
        '''
        return self.xml_doc.findtext('display')

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8 :
